import re

from mskit import rapid_kit
from mskit.constants.mass import CompoundMass, Mass


def calc_prec_mz(pep: str, charge: int = None, mod=None) -> float:
    """
    Example:
    pep = 'LGRPSLSSEVGVIICDISNPASLDEMAK'
    charge = 3
    mod = '15,Carbamidomethyl;26,Oxidation;'
    -> 992.1668502694666

    calc_prec_mz('_VQISPDS[Phospho (STY)]GGLPER_.2')
    -> 717.83486304735
    calc_prec_mz('_TPPRDLPT[Phospho (STY)]IPGVTSPSSDEPPM[Oxidation (M)]EAS[Phospho (STY)]QSHLRNSPEDK_.4')
    -> 1012.2026098813

    :param pep: peptide that is modified with num, e.g. ACDM1M, where 1 equals to M[Oxidation]
    :param charge: precursor charge
    :param mod: str like 1,Carbamidomethyl;3,Oxidation; or list like [(1, 'Carbamidomethyl'), (3, 'Oxidation')]
    :return: Precursor m/z

    TODO 直接给出多少个哪种修饰，或带位点的修饰，或肽段直接带有修饰（需要哪个参数说明）
    """
    if '.' in pep:
        pep, charge = rapid_kit.split_prec(pep)
    if '[' in pep:
        stripped_pep = re.sub(r'\[.+?\]', '', pep).replace('_', '')
        mod_in_pep = re.findall(r'\[(.+?) \(.+?\)\]', pep)
    else:
        stripped_pep = pep.replace('_', '')
        mod_in_pep = None

    pep_mass = 0.
    for aa in stripped_pep:
        pep_mass += Mass.ResMass[aa]
    pep_mass += CompoundMass.CompoundMass['H2O']
    pep_mass += Mass.ProtonMass * charge

    if mod:
        if isinstance(mod, str):
            mod = [_.split(',') for _ in mod.strip(';').split(';')]
        for _each_mod in list(zip(*mod))[1]:
            pep_mass += Mass.ModMass[_each_mod] if _each_mod != 'Carbamidomethyl' else 0.
    if mod_in_pep:
        for each_mod in mod_in_pep:
            if 'Carbamidomethyl' in each_mod:
                continue
            else:
                pep_mass += Mass.ModMass[each_mod]
    return pep_mass / charge


def calc_fragment_mz(pep, frag_type, frag_num, frag_charge, loss_type=None) -> float:
    """
    Example:
    calc_fragment_mz('_VIHDNFGIVEGLM[Oxidation (M)]TTVHAITAT[Phospho (STY)]QK_', 'y', 16, 1, '1,H3PO4')
    -> 1697.8890815221002

    calc_fragment_mz('_[Acetyl (Protein N-term)]SGSS[Phospho (STY)]SVAAMKK_', 'y', 8, 1, '1,H3PO4')
    -> 803.4443855000001

    calc_fragment_mz('_[Acetyl (Protein N-term)]SGSS[Phospho (STY)]SVAAMKK_', 'y', 10, 2, )
    -> 523.24102464735

    calc_fragment_mz('_[Acetyl (Protein N-term)]SGSS[Phospho (STY)]SVAAMKK_', 'b', 6, 1, '1,H3PO4')
    -> 529.2252678

    pep = 'LGRPSLSSEVGVIICDISNPASLDEMAK'
    frag_type = 'y'
    frag_num = 10
    frag_charge = 1
    mod = '15,Carbamidomethyl;26,Oxidation;'
    -> 1091.5037505084001

    :param pep: peptide sequence
    :param frag_type: support b and y ion
    :param frag_num:
    :param frag_charge:
    :param loss_type:
    :return: Fragment m/z
    """

    # TODO 重复使用已经得到的 pep 信息
    stripped_pep, mod_sites, mods = rapid_kit.substring_finder(pep.replace('_', ''))
    mods = [mod.strip('[]()').split(' ')[0] for mod in mods]
    mod_sites = list(map(int, mod_sites))
    frag_num = int(frag_num)
    frag_charge = int(frag_charge)
    frag_mass = Mass.ProtonMass * frag_charge
    if frag_type == 'b':
        mod_dict = dict(zip(mod_sites, mods))
        if 0 in mod_dict:
            frag_mass += Mass.ModMass[mod_dict[0]]

    elif frag_type == 'y':
        stripped_pep = stripped_pep[::-1]
        frag_mass += CompoundMass.CompoundMass['H2O']
        pep_len = len(stripped_pep)
        mod_sites = [pep_len - site + 1 for site in mod_sites]
        mod_dict = dict(zip(mod_sites, mods))
        if frag_num >= pep_len and (pep_len + 1) in mod_sites:
            frag_mass += Mass.ModMass[mod_dict[pep_len + 1]]
    else:
        raise NameError('Only b and y ion are supported')
    for i in range(frag_num):
        frag_mass += Mass.ResMass[stripped_pep[i]]
        if i + 1 in mod_dict:
            mod_type = mod_dict[i + 1]
            if 'Carbamidomethyl' in mod_type:
                continue
            frag_mass += Mass.ModMass[mod_type]

    if loss_type is None or loss_type.lower() == 'noloss':
        pass
    else:
        for loss_num, loss_compound in [loss.split(',') for loss in loss_type.split(';')]:
            frag_mass -= int(loss_num) * Mass.ModLossMass[loss_compound]

    return frag_mass / frag_charge


def calc_fragment_mz_old(pep, frag_type, frag_num, frag_charge, mod=None) -> float:
    """
    Example:
    pep = 'LGRPSLSSEVGVIICDISNPASLDEMAK'
    frag_type = 'y'
    frag_num = 10
    frag_charge = 1
    mod = '15,Carbamidomethyl;26,Oxidation;'
    -> 1091.5037505084001

    :param pep: peptide sequence
    :param frag_type: support b and y ion
    :param frag_num:
    :param frag_charge:
    :param mod:
    :return: Fragment m/z
    """
    frag_num = int(frag_num)
    frag_charge = int(frag_charge)
    frag_mass = Mass.ProtonMass * frag_charge

    if mod:
        if isinstance(mod, str):
            mod = [_.split(',') for _ in mod.strip(';').split(';')]
            mod = [(int(_[0]), _[1]) for _ in mod]
        mod_dict = dict(mod)
    else:
        mod_dict = dict()

    if frag_type == 'b':
        for i in range(frag_num):
            frag_mass += Mass.ResMass[pep[i]]
            if i + 1 in mod_dict:
                frag_mass += Mass.ModMass[mod_dict[i + 1]]
        if 0 in mod_dict:
            frag_mass += Mass.ModMass[mod_dict[0]]

    elif frag_type == 'y':
        frag_mass += CompoundMass.CompoundMass['H2O']
        pep_len = len(pep)
        for i in range(pep_len - 1, pep_len - 1 - frag_num, -1):
            frag_mass += Mass.ResMass[pep[i]]
            if i + 1 in mod_dict:
                frag_mass += Mass.ModMass[mod_dict[i + 1]]
        if frag_num == pep_len:
            frag_mass += Mass.ModMass[mod_dict[0]]
    else:
        raise NameError('Only b and y ion are supported')
    return frag_mass / frag_charge


def get_fragment_mz_dict(pep, fragments, mod=None):
    """
    :param pep:
    :param fragments:
    :param mod:
    :return:
    """
    mz_dict = dict()
    for each_fragment in fragments:
        frag_type, frag_num, frag_charge = rapid_kit.split_fragment_name(each_fragment)
        mz_dict[each_fragment] = calc_fragment_mz(
            pep, frag_type, frag_num, frag_charge, mod)
    return mz_dict



import numpy as np

__all__ = [
    'hex_to_rgb',
    'rgb_to_hex',
    'hsl_to_rgb',
    'rgb_to_hsl',
    'hue_to_rgb',
]


def hex_to_rgb(hex_color, base=256):
    return tuple(int(hex_color[i:i + 2], 16) for i in (1, 3, 5))


def rgb_to_hex(rgb, rgb_base=255, with_sharp=True, upper=True):
    # Check channel num to be 3
    channel_num = len(rgb)
    if channel_num != 3:
        raise ValueError(f'The input rgb value {rgb} need three channels, now has {channel_num}')

    # Trans each rgb channel to 255 based number
    if rgb_base != 255:
        rgb = [int(one_channel * 255 / rgb_base) for one_channel in rgb]

    # Trans to hex code and check if length of hex code is 1
    hex_list = [hex(one_channel)[2:] for one_channel in rgb]
    for i in range(channel_num):
        if len(hex_list[i]) == 1:
            hex_list[i] = f'0{hex_list[i]}'

    # Join hex codes to full hex color
    hex_code = ''.join(hex_list)
    if with_sharp:
        hex_code = '#' + hex_code
    if upper:
        hex_code = hex_code.upper()
    return hex_code


def hsl_to_rgb(hsl, h_type='degree'):
    h, s, l = hsl
    if h_type == 'degree':
        h /= 360
    if hsl[1] == 0:
        rgb = [_ * 255 for _ in hsl]
    else:
        if l < 0.5:
            value_2 = l * (s + 1)
        else:
            value_2 = (s + l) - s * l
        value_1 = 2 * l - value_2
        rgb = [255 * hue_to_rgb(value_1, value_2, h + _ / 3) for _ in range(1, -2, -1)]
    for i, one_rgb in enumerate(rgb):
        if one_rgb % 1 > 0.999:
            rgb[i] += 1
    return [int(_) for _ in rgb]


def rgb_to_hsl(x, h_type='degree'):
    norm_rgb = np.array(x) / 255
    min_value = norm_rgb.min()
    max_value = norm_rgb.max()
    delta_value = max_value - min_value
    l = (min_value + max_value) / 2
    if delta_value == 0:
        h = 0
        s = 0
    else:
        if l < 0.5:
            s = delta_value / (min_value + max_value)
        else:
            s = delta_value / (2 - min_value - max_value)
        delta_rgb = (((max_value - norm_rgb) / 6) + (delta_value / 2)) / delta_value

        min_channel_idx = norm_rgb.argmin()
        sec_calc_index = min_channel_idx + 1 if min_channel_idx < 2 else 0
        h = min_channel_idx / 3 + delta_rgb[min_channel_idx - 1] - delta_rgb[sec_calc_index]
        if h < 0:
            h += 1
        elif h > 1:
            h -= 1
    if h_type == 'degree':
        return h * 360, s, l
    elif h_type == 'decimal':
        return h, s, l
    else:
        raise


def hue_to_rgb(value_1, value_2, hue):
    if hue < 0:
        hue += 1
    elif hue > 1:
        hue -= 1
    if hue < 1 / 6:
        return value_1 + (value_2 - value_1) * 6 * hue
    elif hue < 1 / 2:
        return value_2
    elif hue < 2 / 3:
        return value_1 + (value_2 - value_1) * (2 / 3 - hue) * 6
    else:
        return value_1


def rgb_to_hue():
    pass


def trans_color_code(input_=None, from_='', to_='', *args):
    pass



import re

import numpy as np
import pandas as pd

from mskit import rapid_kit as rk


def substract_snlib(snlib_1, snlib_2, col='Prec'):
    """
    将第一个 lib 按给定的 col 删除第二个 lib 中的内容
    这里两个 lib 的大小不一定哪边一定大，只是删除 overlap 的部分
    """
    pass


def add_pep_pos_in_prot(x, seq_dict: dict):
    """
    如果在一个 protein 匹配到了多个位置，用 / 分隔 （result 中的 PEP.PeptidePosition 是用 , 分隔在同一蛋白多次匹配的情况）
    如果一个 protein group 里有多个蛋白，用 ; 分隔
    如果 protein group 中所有 protein 都没有匹配到，则返回 na
    如果 protein group 中有部分匹配到，则没有匹配到的部分为空 ''，但不影响其他匹配到的位置，如 '54;;' 中的后两个蛋白均没有匹配到
    """
    pep_pos = []
    pep = x['StrippedPeptide']
    if pd.isna(x['ProteinGroups']):  # 如果 result 中找到的肽段没有 assign 蛋白
        return np.nan
    for protein in x['ProteinGroups'].split(';'):
        if protein not in seq_dict:
            pep_pos.append('')
        else:
            protein_seq = seq_dict[protein]
            sites = ''
            for site_index, re_result in enumerate(re.finditer(pep, protein_seq)):
                sites += f'{str(re_result.start() + 1)}/'
            pep_pos.append(sites.strip('/'))
    str_pep_pos = ';'.join(pep_pos)
    if str_pep_pos.strip(';') == '':
        str_pep_pos = np.nan
    return str_pep_pos


def add_phossite_in_pep(x):
    """
    如果一条肽段有多个修饰位点，用 ; 分隔
    """
    if x['IsPhosPep']:
        pep = x['ModifiedPeptide'].replace('_', '')
        sites = []
        for mod_pos, mod in zip(*rk.substring_finder(pep)[1:]):
            if mod == '[Phospho (STY)]':
                sites.append(mod_pos)
        if sites:
            return ';'.join(map(str, sites))
        else:
            return np.nan
    else:
        return np.nan


def add_phossite_in_prot(x):
    """
    如果一条肽段分配给的 protein group 有多个蛋白，用 ; 分隔
    如果一条肽段有多个位点，用 , 分隔
    蛋白 id 和位点间用 - 连接
    如果一条肽段可以匹配到一个蛋白的多个位置，用 / 分隔
    如：P1-142,P1-145/P1-275,P1-278;P2-51,P2-54
        这条肽段的 protein groups 为两个蛋白 P1 和 P2
        这条肽段在 P1 上可以同时作为第 140 位氨基酸开始或 273 位氨基酸开始，其中磷酸化位点在肽段上为 3 位和 6 位，则蛋白序列上的磷酸化位点可以为 142+145 或 275+278 两种情况
        而在 P2 这个蛋白的序列中，这条肽段只出现了一次，肽段在蛋白序列的位置是从 49 开始，则两个位点在蛋白的位置为 51 和 54
    """
    if pd.isna(x['PepPhosSite']) or pd.isna(x['PepPos']) or pd.isna(x['ProteinGroups']):
        return np.nan
    pep_phos_site = x['PepPhosSite'].split(';')
    prot_site = ''
    try:
        for protein, pep_pos in zip(x['ProteinGroups'].split(';'), x['PepPos'].split(';')):  # 如果一个 protein group 有多个 protein
            if pep_pos == '':
                prot_site += ';'
                continue
            for one_pep_pos in pep_pos.split('/'):  # 如果一条肽段匹配到了同个 protein sequence 上的不同位置
                for phos_site in pep_phos_site:  # 如果一条肽段有多个 phosphosite
                    prot_site += f'{protein}-{int(one_pep_pos) + int(phos_site) - 1},'
                prot_site = prot_site.strip(',')
                prot_site += '/'
            prot_site = prot_site.strip('/')
            prot_site += ';'
        prot_site = prot_site[:-1]
    except:
        print(pep_pos, phos_site)
        print(x['Precursor'])
        raise
    return prot_site



import re

from mskit.rapid_kit import substring_finder


def extract_mod_for_deepms2(mod_pep, mod_trans_dict: dict):
    """
    From SN modpep to DeepMS2 modification col needed format
    For mods:
        Acetyl is not considered
        Carbamidomethyl will be ignored
    """
    mod_pep = mod_pep.replace('_', '')
    if '[' not in mod_pep:
        return ''
    else:
        strip_pep, mod_pos, mods = substring_finder(mod_pep)

        mod_info = []
        for pos, mod in zip(mod_pos, mods):
            if 'Carbamidomethyl' in mod:
                continue
            mod_aa = strip_pep[pos - 1]
            mod_name = mod_trans_dict[mod]
            mod_info.append(f'{mod_aa}{pos}{mod_name}')
        return ';'.join(mod_info)


def read_deepms2_phos_result(file, aa_mod_trans_dict: dict):
    """
    From DeepMS2-Phospho to spectra dict with key-value list {int_prec: {spec_dict}, ...}
    Use the .msp file processed by the R script but not the raw .json pred result

    columns = ['StrippedPeptide', 'PrecCharge', 'IntPrec', 'ModInfo', 'IntenDict']
    df =  pd.DataFrame(results, columns=columns)
    df['IntenDict'] = df['IntenDict'].apply(
        lambda x: {frag: inten for frag, inten in x.items() if re.findall('[by](\d+)\+', frag)[0] not in ['1', '2']})
    df['IntenDict'] = df['IntenDict'].apply(mskit.calc.normalize_intensity, max_num=100)
    """
    pred_results = []
    one_spec = dict()
    with open(file, 'r') as f:
        for row in f:
            row = row.strip('\n')

            if row.startswith('Name'):
                pep, prec_charge = row.replace('Name: ', '').split('/')
            elif row.startswith('Comment'):
                mods = row.split('Mods=')[1]
                if mods == '0':
                    mod_pep = pep
                else:
                    split_mods = [mod.split(',') for mod in mods.split('/')][1:]
                    split_mods = [mod for mod in split_mods if mod[2] != 'Carbamidomethyl']
                    if len(split_mods) == 0:
                        mod_pep = pep
                    else:
                        split_mods = [[int(x[0]), *x[1:]] for x in split_mods]
                        split_mods = sorted(split_mods, key=lambda x: x[0])
                        split_pep_seq = list(pep)

                        for mod_idx, mod_aa, mod_name in split_mods:
                            split_pep_seq[mod_idx] = aa_mod_trans_dict[mod_name][mod_aa]
                        mod_pep = ''.join(split_pep_seq)
                int_prec = f'@{mod_pep}.{prec_charge}'
            elif row.startswith('Num peaks'):
                pass
            elif not row:
                if one_spec:
                    pred_results.append([pep, prec_charge, int_prec, mods, one_spec.copy()])
                    one_spec = dict()
                else:
                    pass
            else:
                frag_mz, frag_inten, frag_name = row.split('\t')
                frag_type, frag_num, frag_lossnum, frag_charge = re.findall(r'([by])(\d+?)(-\d+?)?\^(\d)', frag_name)[0]
                if frag_lossnum == '':
                    frag_losstype = 'Noloss'
                elif frag_lossnum == '-17':
                    frag_losstype = 'NH3'
                elif frag_lossnum == '-18':
                    frag_losstype = 'H2O'
                elif frag_lossnum == '-98':
                    frag_losstype = '1,H3PO4'
                else:
                    raise
                trans_frag_name = f'{frag_type}{frag_num}+{frag_charge}-{frag_losstype}'
                one_spec[trans_frag_name] = float(frag_inten)
    return pred_results



"""
In class Enzyme

Enzymes:
    The dict contains all usable enzymes
    Key: The name of enzyme, which starts with uppercase letter except some conventions
    Values:
        CleavageSite: a tuple with three elements, ('cleavage aa', 'direction', 'exclusion')
            'cleavage aa': The AA for cleavage, one AA with one letter representation
            'direction': N-term or C-term of the cleavage AA, -1 and 1
            'exclusion': The AA that makes enzyme miss the target AA
                This will have three situation: None, one AA, and multi AA in one string (None, 'P', 'EPQ')

digestion_site:
    Get AA pairs of the input enzyme.
    For example:
        digestion_site('Trypsin') -> ['KA', 'KC', 'KD', ..., 'RY'] with no 'KP' or 'RP' in it

"""
from .aa import AA


class Enzyme:
    """
    The RE_AllEnd indicates the all possible end pos of the cleavaged peptides, in which the pos of pep end will also be concluded
    """
    Enzymes = {
        'Trypsin': {
            'CleavageSite': [('K', 1, 'P'), ('R', 1, 'P')],
            'RE': '[KR](?!P)',
            'RE_AllEnd': '.*?[KR](?!P)|.+',
        },
        'Trypsin/P': {
            'CleavageSite': [('K', 1, None), ('R', 1, None)],
            'RE': '[KR]',
            'RE_AllEnd': '.*?[KR]|.+',
        },
        'LysC': {
            'CleavageSite': [('K', 1, 'P')],
            'RE': 'K(?!P)',
            'RE_AllEnd': '.*?K(?!P)|.+',
        },
        'LysC/P': {
            'CleavageSite': [('K', 1, 'P')],
            'RE': 'K',
            'RE_AllEnd': '.*?K|.+',
        },
        'LysN': {
            'CleavageSite': [('K', -1, None)],
        },
        'AspC': {
            'CleavageSite': [('D', 1, None)],
        },
        'AspN': {
            'CleavageSite': [('D', -1, None)],
        },
        'Chymotrypsin': {
            'CleavageSite': [('F', 1, None), ('W', 1, None), ('Y', 1, None)],
        },
    }

    AllEnzymes = set(Enzymes.keys()) | set([e.lower() for e in Enzymes.keys()])

    def digestion_site(self, enzyme):
        e = self.Enzymes[enzyme]
        site_info = e['CleavageSite']
        aa_pair = []
        for _info in site_info:
            target_aa = _info[0]
            _direction = _info[1]
            exclusion = _info[2]
            if exclusion:
                accompanied_aa = [_ for _ in AA.AAList_20 if _ not in exclusion]
            else:
                accompanied_aa = AA.AAList_20
            for _acc in accompanied_aa:
                if _direction == 1:
                    _pair = f'{target_aa}{_acc}'
                elif _direction == -1:
                    _pair = f'{_acc}{target_aa}'
                else:
                    raise
                aa_pair.append(_pair)
        return aa_pair

    def __call__(self, enzyme):
        return self.digestion_site(enzyme)



import numpy as np
import scipy.stats

"""
From irt-kit-reference-sheet.xls
"""

Biognosys_iRT11 = """\
standard	Q1	iRT
LGGNEQVTR	487.257	-24.92
GAGSSEPVTGLDAK	644.823	0.00
VEATFGVDESNAK	683.828	12.39
YILAGVENSK	547.298	19.79
TPVISGGPYEYR	669.838	28.71
TPVITGAPYEYR	683.854	33.38
DGLDAASYYAPVR	699.339	42.26
ADVTPADFSEWSK	726.836	54.62
GTFIIDPGGVIR	622.854	70.52
GTFIIDPAAVIR	636.869	87.23
LFLQFGAQGSPFLK	776.93	100.00\
"""

Biognosys_iRT11_Transitions = """\
Q1 monoisotopic	Q1 average	Q3	relative intensity (approximate, TSQ-Vantage)	rank	precursor charge	fragment type	fragment charge	fragment number	nominal sequence	sequence id	transition id
487.257	487.526	860.423	100	1	2	y	1	8	LGGNEQVTR	iRT Kit_a	LGGNEQVTR.y8.1+
487.257	487.526	503.294	53	2	2	y	1	4	LGGNEQVTR	iRT Kit_a	LGGNEQVTR.y4.1+
487.257	487.526	803.401	50	3	2	y	1	7	LGGNEQVTR	iRT Kit_a	LGGNEQVTR.y7.1+
644.822	645.186	800.452	100	1	2	y	1	8	GAGSSEPVTGLDAK	iRT Kit_b	GAGSSEPVTGLDAK.y8.1+
644.822	645.186	604.331	17	2	2	y	1	6	GAGSSEPVTGLDAK	iRT Kit_b	GAGSSEPVTGLDAK.y6.1+
644.822	645.186	1016.53	16	3	2	y	1	10	GAGSSEPVTGLDAK	iRT Kit_b	GAGSSEPVTGLDAK.y10.1+
683.827	684.22	819.389	100	1	2	b	1	8	VEATFGVDESNAK	iRT Kit_c	VEATFGVDESNAK.b8.1+
683.827	684.22	966.453	56	2	2	y	1	9	VEATFGVDESNAK	iRT Kit_c	VEATFGVDESNAK.y9.1+
683.827	684.22	663.295	46	3	2	y	1	6	VEATFGVDESNAK	iRT Kit_c	VEATFGVDESNAK.y6.1+
547.297	547.62	817.442	100	1	2	y	1	8	YILAGVENSK	iRT Kit_d	YILAGVENSK.y8.1+
547.297	547.62	633.321	54	2	2	y	1	6	YILAGVENSK	iRT Kit_d	YILAGVENSK.y6.1+
547.297	547.62	704.358	52	3	2	y	1	7	YILAGVENSK	iRT Kit_d	YILAGVENSK.y7.1+
669.838	670.237	928.416	100	1	2	y	1	8	TPVISGGPYEYR	iRT Kit_e	TPVISGGPYEYR.y8.1+
669.838	670.237	1041.5	48	2	2	y	1	9	TPVISGGPYEYR	iRT Kit_e	TPVISGGPYEYR.y9.1+
669.838	670.237	841.384	39	3	2	y	1	7	TPVISGGPYEYR	iRT Kit_e	TPVISGGPYEYR.y7.1+
683.853	684.264	956.448	100	1	2	y	1	8	TPVITGAPYEYR	iRT Kit_f	TPVITGAPYEYR.y8.1+
683.853	684.264	855.4	59	2	2	y	1	7	TPVITGAPYEYR	iRT Kit_f	TPVITGAPYEYR.y7.1+
683.853	684.264	1069.53	44	3	2	y	1	9	TPVITGAPYEYR	iRT Kit_f	TPVITGAPYEYR.y9.1+
699.338	699.749	855.436	100	1	2	y	1	7	DGLDAASYYAPVR	iRT Kit_g	DGLDAASYYAPVR.y7.1+
699.338	699.749	926.474	68	2	2	y	1	8	DGLDAASYYAPVR	iRT Kit_g	DGLDAASYYAPVR.y8.1+
699.338	699.749	605.341	58	3	2	y	1	5	DGLDAASYYAPVR	iRT Kit_g	DGLDAASYYAPVR.y5.1+
726.835	727.265	1066.48	100	1	2	y	1	9	ADVTPADFSEWSK	iRT Kit_h	ADVTPADFSEWSK.y9.1+
726.835	727.265	533.746	65	2	2	y	2	9	ADVTPADFSEWSK	iRT Kit_h	ADVTPADFSEWSK.y9.2+
726.835	727.265	584.27	23	3	2	y	2	10	ADVTPADFSEWSK	iRT Kit_h	ADVTPADFSEWSK.y10.2+
622.853	623.225	598.368	100	1	2	y	1	6	GTFIIDPGGVIR	iRT Kit_i	GTFIIDPGGVIR.y6.1+
622.853	623.225	713.395	30	2	2	y	1	7	GTFIIDPGGVIR	iRT Kit_i	GTFIIDPGGVIR.y7.1+
622.853	623.225	826.479	28	3	2	y	1	8	GTFIIDPGGVIR	iRT Kit_i	GTFIIDPGGVIR.y8.1+
636.869	637.251	626.399	100	1	2	y	1	6	GTFIIDPAAVIR	iRT Kit_k	GTFIIDPAAVIR.y6.1+
636.869	637.251	854.51	66	2	2	y	1	8	GTFIIDPAAVIR	iRT Kit_k	GTFIIDPAAVIR.y8.1+
636.869	637.251	741.426	60	3	2	y	1	7	GTFIIDPAAVIR	iRT Kit_k	GTFIIDPAAVIR.y7.1+
776.929	777.412	904.489	100	1	2	y	1	9	LFLQFGAQGSPFLK	iRT Kit_l	LFLQFGAQGSPFLK.y9.1+
776.929	777.412	1051.56	91	2	2	y	1	10	LFLQFGAQGSPFLK	iRT Kit_l	LFLQFGAQGSPFLK.y10.1+
776.929	777.412	504.319	80	3	2	y	1	4	LFLQFGAQGSPFLK	iRT Kit_l	LFLQFGAQGSPFLK.y4.1+\
"""




import numpy as np
from matplotlib import pyplot as plt

from mskit.plot_kit.drawing_area.ax_utls import set_thousand_separate


def pcc_hist(pcc_list,
             bin_num=18, bin_color='#696969', bin_alpha=0.75,
             median_line=True, median_color='#009ACD', median_width=3,
             percent_list=(90, 80, 70),
             label_fontsize=8,
             anno_fontsize=8, anno_y_site=(0.75, 0.5, 0.4),
             anno_metric_name='PCC',
             x_lim=(0, 1),
             data_warning=False,
             x_label='PCC', y_label="Number of test peptide precursors", title=None,
             ax=None, save=None):
    # TODO Other metric like SA in annotation
    # TODO Value of metric with certain percentage

    # TODO Check data (e.g. NA values)
    if ax is None:
        ax = plt.gca()

    pcc_list = np.array(pcc_list)
    median_pcc = np.median(pcc_list)
    pcc_num = len(pcc_list)

    set_thousand_separate(ax, axis=('y',))

    ax.set_xlim(*x_lim)
    ax.set_xticks(np.linspace(0, 1, 6))

    hist = ax.hist(pcc_list, bins=bin_num, alpha=bin_alpha, color=bin_color, edgecolor='k')

    if median_line:
        mid_line = ax.axvline(median_pcc, linewidth=median_width, color=median_color)
    else:
        mid_line = None

    ax.set_xlabel(x_label, fontsize=label_fontsize)
    ax.set_ylabel(y_label, fontsize=label_fontsize)

    percent_list = sorted(map(lambda x: x / 100, percent_list))
    pcc_percent_list = []
    anno_title = f'>{anno_metric_name} Percentage'
    pcc_anno_list = [anno_title]
    for pcc in percent_list:
        ratio = np.sum(pcc_list > pcc) / pcc_num
        pcc_percent_list.append(ratio)
        pcc_anno_list.append(f'>{pcc:.2f}  {ratio:.1%}')

    ax.annotate('\n'.join(pcc_anno_list), xy=(0.05, ax.get_ylim()[1] * anno_y_site[0]), fontsize=anno_fontsize)
    ax.annotate('Median {}: {:.3f}'.format(anno_metric_name, median_pcc),
                xy=(0.05, ax.get_ylim()[1] * anno_y_site[1]), fontsize=anno_fontsize)
    ax.annotate(r'n = {:,}'.format(pcc_num), xy=(0.05, ax.get_ylim()[1] * anno_y_site[2]), fontsize=anno_fontsize)

    if title:
        ax.set_title(title)

    if save:
        plt.savefig(save + f'.{anno_metric_name}.png')

    return hist, mid_line, pcc_anno_list



"""

"""

DIANN_OutputDtype = {
    # File
    'File.Name': 'str',
    'Run': 'str',

    # Protein
    'Protein.Group': 'str',
    'Protein.Ids': 'str',
    'Protein.Names': 'object',
    'PG.Quantity': 'float64',
    'PG.Normalised': 'float64',
    'PG.MaxLFQ': 'float64',
    'Protein.Q.Value': 'float64',
    'PG.Q.Value': 'float64',
    'Global.PG.Q.Value': 'float64',
    'First.Protein.Description': 'object',
    'Lib.PG.Q.Value': 'float64',

    # Gene
    'Genes': 'object',
    'Genes.Quantity': 'float64',
    'Genes.Normalised': 'float64',
    'Genes.MaxLFQ': 'float64',
    'Genes.MaxLFQ.Unique': 'float64',
    'GG.Q.Value': 'float64',

    # Peptide
    'Modified.Sequence': 'str',
    'Stripped.Sequence': 'str',
    'Precursor.Id': 'str',
    'Precursor.Charge': 'int',
    'Proteotypic': 'int64',
    'Precursor.Quantity': 'float64',
    'Precursor.Normalised': 'float64',
    'Precursor.Translated': 'float64',

    # Stats & Scores
    'Q.Value': 'float64',
    'PEP': 'float64',
    'Global.Q.Value': 'float64',
    'Translated.Q.Value': 'float64',
    'Lib.Q.Value': 'float64',
    'Mass.Evidence': 'float64',
    'Evidence': 'float64',
    'CScore': 'float64',
    'Decoy.Evidence': 'float64',
    'Decoy.CScore': 'float64',

    # Spec
    'Ms1.Profile.Corr': 'float64',
    'Spectrum.Similarity': 'float64',
    'Fragment.Quant.Raw': 'object',
    'Fragment.Quant.Corrected': 'object',
    'Fragment.Correlations': 'object',
    'MS2.Scan': 'int64',

    'Ms1.Translated': 'float64',
    'Quantity.Quality': 'float64',
    'Ms1.Area': 'float64',

    # PTM
    'PTM.Informative': 'float64',
    'PTM.Specific': 'float64',
    'PTM.Localising': 'float64',
    'PTM.Q.Value': 'float64',
    'PTM.Site.Confidence': 'float64',
    'Lib.PTM.Site.Confidence': 'float64',

    # RT
    'RT': 'float64',
    'iRT': 'float64',
    'RT.Start': 'float64',
    'RT.Stop': 'float64',
    'Predicted.RT': 'float64',
    'Predicted.iRT': 'float64',

    # IM
    'IM': 'float64',
    'iIM': 'float64',
    'Predicted.IM': 'float64',
    'Predicted.iIM': 'float64',
}

DIANN_TsvLibDtype = {

}



from ftplib import FTP

import pandas as pd
import requests

from mskit.constants import ProteomeXchange as PXConst


class PrideFTPDownloader(object):
    def __init__(self, project_id=None):
        self._id = project_id

        self.FileListUrl = None
        self.filelist = None

        self.fileinfo_dict = None
        self.fileinfo_df = None

        self.excluded_files = None
        self._ftp = FTP()

    def get_project_id(self):
        return self._id

    def set_project_id(self, project_id):
        self._id = project_id

    project_id = property(get_project_id, set_project_id, doc='''
    Project ID of ProteomeXchange, like PXD004732''')

    def set_ftp_debuglevel(self, level):
        self._ftp.set_debuglevel(level)

    def get_file_info(self):
        self.FileListUrl = PXConst.Pride.ListQueryUrl + self._id
        request = requests.get(self.FileListUrl)
        if request.status_code != 200:
            raise requests.exceptions.ConnectionError(request.status_code)
        request_json = request.json()
        raw_filelist = request_json['list']
        self.fileinfo_dict = {_['fileName']: _ for _ in raw_filelist}
        self.fileinfo_df = pd.DataFrame(raw_filelist)
        self.filelist = self.fileinfo_df['fileName'].tolist()
        return self.fileinfo_dict

    def get_filelist(self):
        return self.filelist

    def get_exclusion(self):
        return self.excluded_files

    def set_exclusion(self, exclusion):
        self.excluded_files = exclusion

    exclusion = property(get_exclusion, set_exclusion, doc='''''')

    def link_ftp(self):
        try:
            welcome_message = self._ftp.connect(PXConst.Pride.FTP_IP)
        except TimeoutError:
            raise TimeoutError(f'Time out when connect to {PXConst.Pride.FTP_IP}')
        else:
            if '220' not in welcome_message or welcome_message != '220-Welcome to ftp.ebi.ac.uk\n220 ':
                print(f'Connected to ftp server but no correct connect code: {welcome_message}')

    def login_ftp(self, email=None, user='anonymous', password=None):
        if not password:
            if email:
                password = email
            else:
                password = 'your@email.address'
        try:
            login_message = self._ftp.login(user, password)
        except Exception:
            print(f'Login error when login with user=\'{user}\', password=\'{password}\'')
            raise
        else:
            if '230' not in login_message or login_message != '230 Login successful.':
                print(f'Login but no correct login code')

    def ftp_cwd(self):
        pass

    '250 Directory successfully changed.'

    def download(self, ):
        pass

    def __call__(self, filename):
        if self.fileinfo_dict:
            if isinstance(filename, str):
                try:
                    return self.fileinfo_dict[filename]
                except KeyError:
                    raise KeyError(f'No such file {filename}')
            else:
                raise TypeError(f'File name illegal {filename}')
        else:
            raise

    def __del__(self):
        try:
            self._ftp.quit()
        except AttributeError:
            self._ftp.close()

    @staticmethod
    def asp_cmd(year, month, project_id, file_name, target_address):
        ascp_cmd = ' '.join([r'ascp',
                             r'-QT -l 300m -P33001 -i',
                             r'asperaweb_id_dsa.openssh',
                             r'prd_ascp@fasp.ebi.ac.uk:pride/data/archive/{year}/{month}/{project_id}/{file_name}'.format(
                                 year=year, month=month, project_id=project_id, file_name=file_name),
                             rf'{target_address}',
                             ])



class BasicpDeepInfo(object):
    pDeepTrainsetTitle = [
        'spec',
        'peptide',
        'modinfo',
        'b',
        'b-NH3',
        'b-H2O',
        'b-ModLoss',
        'y',
        'y-NH3',
        'y-H2O',
        'y-ModLoss']


MOD = {
    'Carbamidomethyl (C)': 'Carbamidomethyl[C]',
    'Oxidation (M)': 'Oxidation[M]',
}



import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from mskit.plot_kit.drawing_area import remove_target_spines
from mskit.plot_kit.drawing_area import set_thousand_separate


def get_exp_ratio_idx(data: pd.Series, exp_ratio):
    data = data.cumsum() / data.sum()
    exp_ident = []
    for ratio in exp_ratio:
        ident = data[data > ratio].index[0]
        exp_ident.append(ident)
    return data, exp_ident


def cumu_bar(data, exp_ratio=(0.8, 0.9, 0.95, 0.99),
             max_ylim=5000, xlim=(6, 53),
             left_axis_pos=5.5, right_axis_pos=53.5, bottom_axis_pos=-8,

             mark_min_max=True,
             bar_width=0.6, bar_color='#CDC5BF',
             mark_bar_width=0.6, mark_bar_color='#8B8682', mark_line_style='dashed',
             cumu_line_color='#EE7600',
             vline_color='#CDC9A5',
             xlabel='Peptide length', ylabel='Number of stripped peptide',
             cumu_ylabel='Cumulation percent',
             title='',
             ax=None):
    if ax is None:
        ax = plt.gca()

    if isinstance(data, list):
        ser = pd.Series(dict(data))
    elif isinstance(data, dict):
        ser = pd.Series(data)
    elif isinstance(data, pd.Series):
        ser = data
    else:
        raise TypeError
    ser = ser.sort_index()

    ax_cumu = ax.twinx()

    remove_target_spines(['top', 'right'], ax=ax)
    remove_target_spines(['top', 'left', 'bottom'], ax=ax_cumu)

    min_data_x, max_data_x = ser.index.min(), ser.index.max()
    min_data_y, max_data_y = ser.min(), ser.max()

    percent_cumu_ser, marked_idx = get_exp_ratio_idx(ser, exp_ratio=exp_ratio)
    cumu_num = percent_cumu_ser * max_data_y

    if max_ylim:
        ...
    else:
        ...

    for idx, num in ser.items():
        if idx in marked_idx:
            ax.bar(idx, num, width=mark_bar_width, bottom=True, color=mark_bar_color)
            ax.vlines(x=idx, ymin=num, ymax=cumu_num.loc[idx],
                      colors=vline_color, linestyles=mark_line_style, label='')
        elif mark_min_max and idx in [min_data_x, max_data_x]:
            ax.bar(idx, num, width=mark_bar_width, bottom=True, color=mark_bar_color)
        else:
            ax.bar(idx, num, width=bar_width, bottom=True, color=bar_color)

    cumu_line = ax_cumu.plot(ser.index.tolist(), cumu_num.values, color=cumu_line_color)

    ax.set_ylim(0, max_ylim)
    ax.set_xlim(*xlim)
    ax.set_xticks([min_data_x, *marked_idx, max_data_x])
    set_thousand_separate(ax=ax, axis='y')

    cumu_ax_yscale = max_data_y
    ax_cumu.set_ylim(bottom=0, top=max_ylim, emit=True, auto=False, ymin=None, ymax=None)
    ax_cumu.set_yticks(np.arange(0, 1.1, 0.1) * cumu_ax_yscale)
    ax_cumu.set_yticklabels(['{}%'.format(_) for _ in range(0, 101, 10)])

    ax.spines['left'].set_position(('data', left_axis_pos))
    ax.spines['bottom'].set_position(('data', bottom_axis_pos))
    ax_cumu.spines['right'].set_position(('data', right_axis_pos))

    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax_cumu.set_ylabel(cumu_ylabel)

    ax.set_title(title)



from .rt_corr import delta_tx, r_square
from .spec_similarity import (
    inten_pair_from_dict, inten_pair_from_array,
    get_frag_list, calc_sa, calc_pcc, calc_pcc_with_dictdata, calc_sa_with_dictdata, frag_sa, frag_pcc
)



import os
import re
from collections import defaultdict
from os.path import join as join_path

import pandas as pd

from mskit import rapid_kit as rk


def collect_plink_results(data_dir):
    used_linkers = os.listdir(data_dir)
    used_linkers = [_ for _ in used_linkers if os.path.isdir(join_path(data_dir, _))]
    plink_results = {'Cross': defaultdict(list), 'Loop': defaultdict(list), 'Mono': defaultdict(list)}
    for linker in used_linkers:
        linker_dir = join_path(data_dir, linker)
        for rep_name in os.listdir(linker_dir):
            rep_dir = join_path(linker_dir, rep_name)
            for file_name in os.listdir(rep_dir):
                file_path = join_path(rep_dir, file_name)
                df = rk.flatten_two_headers_file(file_path, sep=',')
                if df.size == 0:
                    continue
                df['Linker'] = linker
                df['Source'] = f'{linker}/{rep_name}/{file_name}'
                if 'cross-linked' in file_name:
                    df['LinkType'] = 'CrossLink'
                    plink_results['Cross'][linker].append(df.copy())
                elif 'loop-linked' in file_name:
                    df['LinkType'] = 'LoopLink'
                    df['Protein_Type'] = 'Intra-Protein'
                    plink_results['Loop'][linker].append(df.copy())
                elif 'mono-linked' in file_name:
                    df['LinkType'] = 'MonoLink'
                    plink_results['Mono'][linker].append(df.copy())
                else:
                    raise
    for link_type in list(plink_results.keys()):
        for linker in list(plink_results[link_type].keys()):
            plink_results[link_type][linker] = pd.concat(plink_results[link_type][linker]).reset_index().rename({'index': 'RawIndex'}, axis=1)
    return plink_results


def split_cross_link_info(x, prot_name_mapper):
    link_info = x['Proteins']

    re_match = re.match(r'(.+?)\((\d+)\)-(.+?)\((\d+)\).*', link_info)
    prot_site = [(re_match.group(1), int(re_match.group(2))),
                 (re_match.group(3), int(re_match.group(4)))]

    sorted_prot_site = sorted(prot_site, key=lambda _: _[1])
    sorted_link_info = '{}({})-{}({})'.format(prot_name_mapper[sorted_prot_site[0][0]], sorted_prot_site[0][1],
                                              prot_name_mapper[sorted_prot_site[1][0]], sorted_prot_site[1][1])

    pep = x['Peptide']
    if str(sorted_prot_site[0][1]) != re_match.group(2):
        pep = '-'.join(pep.split('-')[::-1])

    return pep, *rk.sum_list(prot_site), sorted_link_info


def process_cross_link_result(df, prot_name_mapper):
    # 非 gi| 的部分 proteins 均为单组位点
    df = df[~df['Proteins'].str.contains('gi|', regex=False)].copy()

    df[[
        'SortedPeptide',
        'Raw_Prot1',
        'Raw_Site1',
        'Raw_Prot2',
        'Raw_Site2',
        'SortedLinkInfo'
    ]] = df.apply(split_cross_link_info, args=(prot_name_mapper,), axis=1, result_type='expand')
    return df.copy()


def split_loop_link_info(x, prot_name_mapper):
    link_info = x['Proteins']

    re_match = re.match(r'(.*?)\((\d+)\)\((\d+)\).*', link_info)
    prot_site = [re_match.group(1), int(re_match.group(2)), int(re_match.group(3))]
    sorted_site = sorted(prot_site[1:])
    sorted_link_info = '{0}({1})-{0}({2})'.format(prot_name_mapper[prot_site[0]], sorted_site[0], sorted_site[1])

    pep = x['Peptide']
    re_match = re.match(r'(.+?)\((\d+)\)\((\d+)\).*', pep)
    pep = '{0}({1})-{0}({2})'.format(
        re_match.group(1),
        *sorted([int(re_match.group(2)), int(re_match.group(3))])
    )

    return pep, *prot_site, sorted_link_info


def process_loop_link_result(df, prot_name_mapper):
    df = df[~df['Proteins'].str.contains('gi|', regex=False)].copy()

    df[[
        'SortedPeptide',
        'Raw_Prot',
        'Raw_Site1',
        'Raw_Site2',
        'SortedLinkInfo'
    ]] = df.apply(split_loop_link_info, args=(prot_name_mapper,), axis=1, result_type='expand')
    return df.copy()



from matplotlib import pyplot as plt


def bar_anno(anno_text, x_pos, y_height=None, y_bottom=0,
             thousand_separ=True, add_percent=False, denominator=None,
             ha='center', va='center',
             rotation=90, fontsize=8,
             ax=None):
    if ax is None:
        ax = plt.gca()

    if thousand_separ is True:
        showed_anno = format(anno_text, ',')
    elif thousand_separ is False:
        showed_anno = str(anno_text)
    else:
        showed_anno = None

    if not y_height:
        y_height = anno_text

    y_pos = y_height / 2 + y_bottom
    ax.annotate(showed_anno, (x_pos, y_pos),
                ha=ha, va=va, rotation=rotation, fontsize=fontsize, )


def dyna_bar_anno(anno_text, x_pos, y_height=None, y_bottom=0,
                  thousand_separ=True,
                  ha='center', va='center',
                  rotation=90, fontsize=8,
                  is_dyna_func=lambda x, y: abs(x) < (abs(y) / 2),
                  dyna_pos_func=lambda x, y: 2 * x,
                  auto_ha_va=False,
                  new_ha=None, new_va=None, new_rotation=None,
                  ax=None):
    """
    如果是本身就很低的一个bar但是bottom是0还是会被挡，太菜了，把 dyna func 写成 lambda x, y: x < 500 之类的海星

    TODO 把 dyna func 这部分拿出来单独处理，放进 unit test中
    TODO 这里同时定义两个 func 也太麻烦了
    """

    if not y_height:
        y_height = anno_text

    is_dyna = is_dyna_func(y_height, y_bottom)

    if is_dyna:
        y_height = dyna_pos_func(y_height, y_bottom)

    if auto_ha_va is True and is_dyna:
        va = 'bottom' if y_height > 0 else 'top'
    elif auto_ha_va is False:
        if new_ha and is_dyna:
            ha = new_ha
        if new_va and is_dyna:
            va = new_va
        if new_rotation and is_dyna:
            rotation = new_rotation
    else:
        pass

    bar_anno(anno_text, x_pos, y_height=y_height, y_bottom=y_bottom,
             thousand_separ=thousand_separ,
             ha=ha, va=va,
             rotation=rotation, fontsize=fontsize,
             ax=ax)



import numpy as np
import pandas as pd


def sum_list(nested_list):
    temp = []
    for _ in nested_list:
        temp.extend(_)
    return temp


flatten_list = sum_list


def recursive_sum_list(nested_list: list) -> list:
    """
    Recursively flatten list consist of both list/tuple and un-iterable items
    TODO append method will take much time?
        # inner_item_is_list = [isinstance(_, _) for _ in l]
        # if any(inner_item_is_list):
    """
    temp = []
    for item in nested_list:
        if isinstance(item, (list, tuple)):
            temp.extend(recursive_sum_list(item))
        else:
            temp.append(item)
    return temp


def sum_set_in_list(list_with_sets, return_type='set'):
    union_set = list_with_sets[0]
    if len(list_with_sets) >= 2:
        for s in list_with_sets[1:]:
            union_set = union_set | s
    if return_type == 'set':
        return union_set
    elif return_type == 'list':
        return list(union_set)
    else:
        raise ValueError('Not supported return type when sum set list')


def align_dict(*dn: dict, columns=None):
    aligned_list = []
    key_union = sorted(set(sum_list(dn)))
    for key in key_union:
        aligned_list.append([di.get(key, np.nan) for di in dn])
    if columns is None:
        columns = [f'D{i}' for i in range(1, len(dn) + 1)]
    return pd.DataFrame(aligned_list, index=key_union, columns=columns)


def split_two_set(set1, set2):
    overlapped = set1 & set2
    set1_unique = set1 - set2
    set2_unique = set2 - set1
    return set1_unique, overlapped, set2_unique


def drop_list_duplicates(initial_list):
    unique_list = list(set(initial_list))
    unique_list = sorted(unique_list, key=initial_list.index)
    return unique_list


def intersect_lists(list_1, list_2, drop_duplicates=True):
    intersected_list = [_ for _ in list_1 if _ in list_2]
    if drop_duplicates:
        return drop_list_duplicates(intersected_list)
    else:
        return intersected_list


def subtract_list(list_1, list_2, drop_duplicates=True):
    subtracted_list = [_ for _ in list_1 if _ not in list_2]
    if drop_duplicates:
        return drop_list_duplicates(subtracted_list)
    else:
        return subtracted_list


def get_dict_coincide_data(dict_1, dict_2):
    shared_keys = list(set(dict_1.keys()) & set(dict_2.keys()))
    value_list_1 = [dict_1[_] for _ in shared_keys]
    value_list_2 = [dict_2[_] for _ in shared_keys]
    return shared_keys, value_list_1, value_list_2


def str_mod_to_list(mod):
    mod_list = [each_mod.split(',') for each_mod in mod.strip(';').split(';')]
    mod_list = [(int(_[0]), _[1]) for _ in mod_list]
    return mod_list


def check_value_len_of_dict(checked_dict: dict, thousands_separator=True, sort_keys=False, ):
    # TODO sort_keys 可以为 lambda 函数
    # TODO return length dict or print
    if sort_keys:
        keys = sorted(checked_dict.keys())
    else:
        keys = checked_dict.keys()
    for k in keys:
        v = checked_dict[k]
        v_len = len(v)
        if thousands_separator:
            print(f'{k}: {format(v_len, ",")}')
        else:
            print(f'{k}: {v_len}')


class XmlListConfig(list):
    def __init__(self, x_list):
        super(XmlListConfig, self).__init__()
        for element in x_list:
            if element:
                if len(element) == 1 or element[0].tag != element[1].tag:
                    self.append(XmlDictConfig(element))
                elif element[0].tag == element[1].tag:
                    self.append(XmlListConfig(element))
            elif element.text:
                text = element.text.strip()
                if text:
                    self.append(text)


class XmlDictConfig(dict):
    def __init__(self, parent_element):
        super(XmlDictConfig, self).__init__()
        if parent_element.items():
            self.update(dict(parent_element.items()))
        for element in parent_element:
            if element:
                if len(element) == 1 or element[0].tag != element[1].tag:
                    x_dict = XmlDictConfig(element)
                else:
                    x_dict = {element[0].tag: XmlListConfig(element)}
                if element.items():
                    x_dict.update(dict(element.items()))
                self.update({element.tag: x_dict})
            elif element.items():
                self.update({element.tag: dict(element.items())})
            else:
                self.update({element.tag: element.text})


def xml_to_dict(xml_context):
    from xml.etree import cElementTree as ElementTree

    _root = ElementTree.XML(xml_context)
    _xml_dict = XmlDictConfig(_root)
    return _xml_dict



import re


class ModProcess(object):
    @staticmethod
    def add_mod(pep: str, mod_type=None):
        """
        Add fixed modification 'C[Carbamidomethyl (C)]'
        Variable modificaion: only M[Oxidation (M)] is supported now, only one oxidation mod will be added when MOD is ['M']
        TODO: Add other modifications and change modification number
        :param pep:
        :param mod_type:
        :return:
        """
        modpep_list = []
        mod_pep = pep.replace('C', MOD.ModDict['C'])

        if mod_type == ['M']:
            if 'M' in mod_pep:
                for _ in re.finditer('M', mod_pep):
                    m_site = _.start()
                    modpep_list.append(mod_pep[:m_site] + MOD.ModDict['M'] + mod_pep[m_site + 1:])
            else:
                modpep_list.append(mod_pep)
        else:
            modpep_list.append(mod_pep)

        return modpep_list

    @staticmethod
    def modpep2intmod():
        """
        C[Carbamidomethyl (C)] -> C[+16]
        int mod pep is not the necessary column in spectronaut library, so it is not used now
        :return:
        """
        pass

    @staticmethod
    def remove_str_mod(mod_pep):
        pep = mod_pep.replace(MOD.ModDict['C'], 'C').replace(MOD.ModDict['M'], 'M')
        return pep

    @staticmethod
    def remove_int_mod(mod_pep):
        pep = mod_pep.replace('1', 'M')
        return pep

    @staticmethod
    def mod2new(old_mod_pep):
        new_mod_pep = old_mod_pep.replace(MOD.ModDict_old['C'], MOD.ModDict_new['C']).replace(MOD.ModDict_old['M'], MOD.ModDict_new['M'])
        return new_mod_pep

    @staticmethod
    def mod2old(new_mod_pep):
        old_mod_pep = new_mod_pep.replace(MOD.ModDict_new['C'], MOD.ModDict_old['C']).replace(MOD.ModDict_new['M'], MOD.ModDict_old['M'])
        return old_mod_pep

    @staticmethod
    def mod_num2str(num_mod_pep):
        str_mod_pep = num_mod_pep.replace('C', MOD.ModDict['C']).replace('1', MOD.ModDict['M'])
        return str_mod_pep

    @staticmethod
    def mod_str2num(str_mod_pep):
        num_mod_pep = str_mod_pep.replace(MOD.ModDict['C'], 'C').replace(MOD.ModDict['M'], '1')
        return num_mod_pep



from . import calc
from . import config
from . import constants
from . import link_db
from . import metric
from . import ms_pred
from . import post_analysis
from . import rapid_kit
from . import sequence
from . import stats

from .inherited_builtins import *



import re

import pandas as pd

"""
检索结果的导入
    1. 每种文件单独
    2. 所有文件在一个文件夹中，读入指定文件关联 id

"""


def mq_modpep_to_intseq_1_6(x):
    x = x.replace('_', '')
    if '(Acetyl (Protein N-term))' in x:
        x = x.replace('(Acetyl (Protein N-term))', '')
        x = '*' + x
    else:
        x = '@' + x
    x = x.replace('C(Carbamidomethyl (C))', 'C')
    x = x.replace('M(Oxidation (M))', '1')
    x = x.replace('S(Phospho (STY))', '2')
    x = x.replace('T(Phospho (STY))', '3')
    x = x.replace('Y(Phospho (STY))', '4')
    return x


def mq_modpep_to_intseq_1_5(x):
    x = x.replace('_', '')
    if '(ac)' in x:
        x = x.replace('(ac)', '')
        x = '*' + x
    else:
        x = '@' + x
    x = x.replace('C(Carbamidomethyl (C))', 'C')
    x = x.replace('M(ox)', '1')
    x = x.replace('S(ph)', '2')
    x = x.replace('T(ph)', '3')
    x = x.replace('Y(ph)', '4')
    return x


def ion_info_from_mq(need_ion_type=('b', 'y'), verify_mass=True, verify_tolerance=20):
    pass


def inten_from_mq(x):
    ions = x['Matches']
    intens = x['Intensities']

    ion_intens_list = list(zip(ions.split(';'), intens.split(';')))

    inten_dict = dict()

    for ion_type in ['b', 'y']:
        ion_info = [_ for _ in ion_intens_list if _[0].startswith(ion_type)]

        current_num = 0
        ex_mod = 0
        for ion, inten in ion_info:

            ion_num = re.findall(f'{ion_type}(\d+)', ion)[0]
            ion_num = int(ion_num)

            re_charge = re.findall('\((\d)\+\)', ion)
            if re_charge:
                ion_charge = re_charge[0]
            else:
                ion_charge = '1'

            frag = f'{ion_type}{ion_num}+{ion_charge}'

            if '-' in ion:
                loss_type = re.findall('-(.+)$', ion)[0]
                frag += f'-1,{loss_type}'
                if ex_mod:
                    frag += f';{ex_mod},H3PO4'

            elif '*' in ion:
                if ex_mod == 0:
                    current_num = 0
                    ex_mod = 1
                else:
                    if ion_num <= current_num:
                        ex_mod += 1
                frag += f'-{ex_mod},H3PO4'

            else:
                frag += f'-Noloss'

            current_num = ion_num

            inten_dict[frag] = float(inten)

    return inten_dict


def show_ion_pair(result_df, iloc, return_df=True):
    ion_pairs = list(zip(result_df.iloc[iloc]['Masses'].split(';'),
                         result_df.iloc[iloc]['Matches'].split(';'),
                         result_df.iloc[iloc]['Intensities'].split(';')))
    if return_df:
        return pd.DataFrame(ion_pairs, columns=['FragMz', 'Frag', 'Inten'])
    else:
        for _ in ion_pairs:
            print(_)


def remove_reverse_hits(evi_msms: pd.DataFrame):
    return evi_msms.loc[evi_msms['Reverse'] != '+']


def remove_zero_inten_hits(evi_msms: pd.DataFrame):
    return evi_msms.loc[evi_msms['Intensity'] > 0]



import re

import numpy as np
import pandas as pd

from .text_kit import substring_finder


def select_row_with_target_col(df, ident, colname, return_col_list=None):
    _df = df[df[colname].str.contains(ident)]
    if return_col_list:
        return _df[return_col_list]
    else:
        return _df


def extract_df_with_col_ident(original_df, identifiers, focus_col, return_col_list=None):
    if isinstance(identifiers, str):
        target_df = select_row_with_target_col(original_df, identifiers, focus_col, return_col_list)
        return target_df
    elif isinstance(identifiers, list):
        target_df_list = [select_row_with_target_col(original_df, _, focus_col, return_col_list) for _ in identifiers]
        return target_df_list
    else:
        return None


def df_keep_block_first_line(df_with_blocks, filtered_col_name):
    """
    To get the first line when the file is consist of many blocks which means some same values are in the same column in neighboring lines
    This may keep same values in the selected columns because the blocks have same values in the selected column may not be neighboring
    :param df_with_blocks:
    :param filtered_col_name:
    :return:
    """
    non_block_df = df_with_blocks[df_with_blocks[filtered_col_name] != df_with_blocks[filtered_col_name].shift(1)]
    return non_block_df


def match_protein_groups(x, protein_list, col='PG.ProteinGroups', delimiter=';'):
    if isinstance(x[col], float):
        return False
    if set(x[col].split(delimiter)) & set(protein_list):
        return True
    return False


def split_protein_groups(x, col=None):
    if not isinstance(x, str):
        acc = x[col]
    else:
        acc = x
    if pd.isna(acc):
        return np.nan
    acc = acc.split(';')[0].strip()
    return acc


def df_split_protein_groups(df, col='PG.ProteinAccessions'):
    first_protein = df.apply(split_protein_groups, col=col, axis=1)
    df['FirstProtein'] = first_protein
    return df


def add_protein_type(x, protein_content, col='FirstProtein'):
    acc = split_protein_groups(x, col=col)
    if pd.isna(acc):
        return np.nan
    for protein_type, acc_list in protein_content.items():
        if acc in acc_list:
            return protein_type
    return 'Others'


def df_add_protein_type(df, col, protein_content, used_type=None, return_col='ProteinType'):
    if isinstance(protein_content, list):
        protein_content = {used_type: protein_content}
    type_series = df.apply(add_protein_type, protein_content=protein_content, col=col, axis=1)
    df[return_col] = type_series
    return df


def filter_prob(
        x,
        find_col,
        prob_col,
        mod_name='(Phospho (STY))',
        recept_prob=0.75,
        refute_prob=0.75):
    find_col_content = x[find_col]
    if mod_name not in find_col_content:
        return True

    prob_col_content = x[prob_col]
    if pd.isna(prob_col_content):
        return True

    if f'(' not in prob_col_content:
        return True

    strippep, modpos, mods = substring_finder(find_col_content.replace('_', ''), '(', ')')
    strippep, prob_pos, probs = substring_finder(prob_col_content.replace('_', ''), '(', ')')

    for pos in prob_pos:
        prob_idx = prob_pos.index(pos)

        prob = float(probs[prob_idx].strip('()'))
        if pos in modpos:
            if prob <= recept_prob:
                return False
        else:
            if prob > refute_prob:
                return False
    return True


def __filter_prob(
        x,
        find_col,
        prob_col,
        ident='ph',
        recept_prob=0.75,
        refute_prob=0.75):
    """
    :param x: one row of the target dataframe
    :param find_col: colname of col contains sequence with determined label. Example: col 'Modified sequence' _(ac)AAAAAAAAAAAAGDS(ph)DS(ph)WDADTFSM(ox)EDPVRK_
    :param prob_col: colname of col contains sequence with probability. Example: col 'Phospho (STY) Probabilities' AAAAAAAAAAAAGDS(0.876)DS(0.887)WDADT(0.161)FS(0.077)MEDPVRK
    """
    # TODO 这里改成所有含 Probability 的列都 filter，单列filter可选

    find_col_content = x[find_col]
    # To avoid error if there is no mod in the seq. 'Modified sequence' will
    # have no Parentheses
    if f'({ident})' not in find_col_content:
        return True
    prob_col_content = x[prob_col]
    # To avoid error if there is no target var mod in the seq. Prob col will
    # be NaN

    if pd.isna(prob_col_content):
        return True

    if f'(' not in prob_col_content:
        return True

    # replace_pattern = re.compile(rf'\([^{ident}]+\)')  # TODO 这里好像是对 p h 两个字母单独 find
    replace_pattern = re.compile(rf'\((?!{ident}).+?\)')
    ident_find_seq = re.sub(
        replace_pattern,
        '',
        find_col_content).replace(
        '_',
        '')
    split_find_col = re.split(r'[\(\)]', ident_find_seq)[::2]
    split_prob_col = re.split(r'[\(\)]', prob_col_content)
    find_seq_sum = [''.join(split_find_col[:i + 1])
                    for i in range(len(split_find_col))]
    prob_seq_sum = ''
    # 这里因为 prob col 中的位点概率会给出所有可能分配位点的概率值
    # 但是 sequence 里没有过 cutoff 的则没有给出修饰信息，即 (ph)
    # 因此通过 prob col 得到的 sequence 可能不在分割出的 mod sequence 列中
    # 这里设置 refute prob 是指没有给出修饰位点但是这个位点的 prob 超过了一定的值
    for i in range(0, len(split_prob_col) - 1, 2):
        prob_seq_sum += split_prob_col[i]
        if prob_seq_sum in find_seq_sum:
            if float(split_prob_col[i + 1]) <= recept_prob:
                return False
        else:
            if float(split_prob_col[i + 1]) > refute_prob:
                return False
    return True



from mskit.plot_kit.colors import PreColorDict
from mskit.plot_kit.drawing_area import ax_utls
from mskit.plot_kit.annotations import bar_anno, dyna_bar_anno

from matplotlib import pyplot as plt


# Basic inten plot -> combine to mirror plot

def mirror_plot(upper_data, lower_data, colors=None, ls=None, lw=1, ):
    pass



import numpy as np
from scipy.spatial.distance import euclidean


def _c(ca, i, j, P, Q):
    if ca[i, j] > -1:
        return ca[i, j]
    elif i == 0 and j == 0:
        ca[i, j] = euclidean(P[0], Q[0])
    elif i > 0 and j == 0:
        ca[i, j] = max(_c(ca, i - 1, 0, P, Q),
                       euclidean(P[i], Q[0]))
    elif i == 0 and j > 0:
        ca[i, j] = max(_c(ca, 0, j - 1, P, Q),
                       euclidean(P[0], Q[j]))
    elif i > 0 and j > 0:
        ca[i, j] = max(
            min(_c(ca, i - 1, j, P, Q),
                _c(ca, i - 1, j - 1, P, Q),
                _c(ca, i, j - 1, P, Q)),
            euclidean(P[i], Q[j]))
    else:
        ca[i, j] = float("inf")
    return ca[i, j]


def frechet_dist(P, Q):
    """
    Discrete Frechet Distance

    Usage
    -----
    Compute the discret frechet distance between trajectories P and Q
    Parameters
    ----------
    param P : px2 numpy_array, Trajectory P
    param Q : qx2 numpy_array, Trajectory Q
    Returns
    -------
    frech : float, the discret frechet distance between trajectories P and Q
    """
    ca = np.ones((len(P), len(Q)))
    ca = np.multiply(ca, -1)
    return _c(ca, len(P) - 1, len(Q) - 1, P, Q)



from mskit.constants import BasicModification
from mskit import rapid_kit


class ModOperation(BasicModification):
    def __init__(self):
        """
        self._mod_rule stores the mod rule for peptide. Default is pre-defined StandardModRule.
        self._mod_rule_dict stores manurally defined extra mod rules
        self._display_mod the mods set for assembling queried mod
        self._mod_aa the possible AA for the target mod
        """
        self._mod_rule = self.StandardModRule
        self._mod_rule_dict = self.ModRuleDict
        self._extra_mod_rule_name_list = []

        self._alias_dict = self.ModAliasDict

        self._display_mod = self.StandardMods

        self._mod_aa = self.ModAA

    def __call__(self, mod=None, aa=None):
        """
        To get the mod directly
        """
        return self.query_mod(mod=mod, aa=aa)

    def query_mod(self, mod=None, aa=None):
        """
        Get the mod for display
        Example:
            If mod rule is default standard rule: mod='Carbamidomethyl', aa='C' -> '[Carbamidomethyl (C)]'
            If mod rule is '({mod})' which is used in MaxQuant. mod='ph' -> '(ph)'
        """
        standard_mod = self.get_standard_mod(mod)
        used_mod = self._display_mod[standard_mod]
        return self._mod_rule.format(mod=used_mod, aa=aa)

    def get_standard_mod(self, modname):
        return self._alias_dict[modname]

    def set_extra_mod_rule(self, extra_mod_rule, rule_name='extra_1', set_used=True):
        """
        :param extra_mod_rule
        This receive a string to define the format of mod rule.
        There are two reserved parameters can be used: mod, aa
        Example: '[{mod} (aa)}]', this will define a mod rule like '[Carbamidomethyl (C)]'
        :param set_used
        True means set the currently used mod rule to this extra mod rule
        :param rule_name
        For storing the name of extra mod rule.
        """
        self._mod_rule_dict[rule_name] = extra_mod_rule
        self._extra_mod_rule_name_list.append(rule_name)
        if set_used:
            self._mod_rule = extra_mod_rule

    def add_mod_alias(self, alias_dict: dict, method='direct'):
        """
        :param alias_dict: A dict to define name aliases of modifications. Details are listed in method param.
        :param method: 'direct' or 'list'.
            'direct' to extend the alias dict directly with key-value pairs like: {'alias_mod': 'standard_mod'}
            'list' to extend the alias dict via multi aliases with format: {'standard_mod': ['alias_1', 'alias_2']}
        """
        if method == 'direct':
            for alias, standard in alias_dict.items():
                self._alias_dict[alias] = self._alias_dict[standard]
        elif method == 'list':
            for standard, aliases in alias_dict.items():
                for alias in aliases:
                    self._alias_dict[alias] = self._alias_dict[standard]
        else:
            raise

    def set_display_mod(self, mod_dict: dict, method='add'):
        """
        Define the mods for display (i.e. the mod string for assembly)
        :param mod_dict
        The key of mod_dict is any mod name that can be found in the alias dict,
        and the key will be converted to standard mod name.
        Then the value will be a new display mod name.
        e.g. mod_dict={'ox': 'oxi'} -> from alias, 'ox' is 'Oxidation', then the mod for mod rule will be 'oxi'
        :param method
        'add' means add the new display mods and the shared mods will be replaced to new ones
        'new' means the new display mod dict will completely overwrite the previous one
        """
        _mod_dict = dict()
        for mod, disp_mod in mod_dict.items():
            standard_mod = self.get_standard_mod(mod)
            _mod_dict[standard_mod] = disp_mod
        if method == 'add':
            self._display_mod = {**self._display_mod, **_mod_dict}
        elif method == 'new':
            self._display_mod = _mod_dict
        else:
            raise

    def extend_mod(self, mod):
        """
        This receive mod to extend the pre-defined mods
        Example:
            mod='NewMod' -> self._display_mod['NewMod'] = 'NewMod' and self._alias_dict['NewMod'] = 'NewMod'
            This will add the mod itself for new query
        :param mod New mod name
        """
        self._display_mod[mod] = mod
        self._alias_dict[mod] = mod

    def switch_mod_rule(self, rule=None):
        if rule == 'standard':
            self._mod_rule = self._mod_rule_dict['standard']
        else:
            if rule:
                try:
                    self._mod_rule = self._mod_rule_dict[rule]
                except KeyError:
                    print('No extra mod rule defined, change to standard mod rule')
                    self._mod_rule = self._mod_rule_dict['standard']
            else:
                rule = self._extra_mod_rule_name_list[-1]
                self._mod_rule = self._mod_rule_dict[rule]

    def show_extra_mod_rule_list(self):
        return self._extra_mod_rule_name_list

    def return_mod_rule(self):
        return self._mod_rule

    def add_mod(self, pep, mod):
        mod_pep = rapid_kit.add_mod(pep=pep, mod=mod, mod_processor=self)
        return mod_pep

    def query_possible_aa(self, mod_name):
        standard_mod = self.get_standard_mod(mod_name)
        return self._mod_aa[standard_mod]

    def __extract_mod(self, mod_seq):
        pass

    def __trans_mod(self, modpep):
        pass



import re
from itertools import combinations

import pandas as pd
from tqdm import tqdm

from mskit import rapid_kit


def extract_seq_window(seq, fasta_dict: dict, n=7, ex_pad_symbol='_'):
    """
    TODO fasta dict 可以传入 fasta parser
    TODO 可以指定蛋白序列 / 指定蛋白 acc
    """
    ext_data = []
    for acc, prot_seq in fasta_dict.items():
        if seq in prot_seq:
            find_pos = [_.span() for _ in re.finditer(seq, prot_seq)]
            for start_idx, end_pos in find_pos:
                if start_idx < n:
                    prev_seq = ex_pad_symbol * (n - start_idx) + prot_seq[: start_idx]
                else:
                    prev_seq = prot_seq[start_idx - n: start_idx]
                back_seq = prot_seq[end_pos: end_pos + n]
                if len(back_seq) < n:
                    back_seq = back_seq + ex_pad_symbol * (n - len(back_seq))
                ext_data.append((seq, prev_seq, back_seq, acc, start_idx + 1))
    return ext_data


def batch_extract_seq_window(seq_list, fasta_dict: dict, n=7, ex_pad_symbol='_', return_type='df'):
    extract_data = []
    not_find_data = []
    with tqdm(seq_list) as t:
        t.set_description(f'Extract seq window (n={n}): ')
        for seq in t:
            result = extract_seq_window(seq, fasta_dict=fasta_dict, n=n, ex_pad_symbol=ex_pad_symbol)
            if result:
                extract_data.extend(result)
            else:
                not_find_data.append(seq)
    if return_type == 'raw':
        return extract_data, not_find_data
    elif return_type == 'df':
        raw_df = pd.DataFrame(extract_data, columns=['Pep', 'Prev', 'Back', 'Acc', 'StartPos'])
        group_df = raw_df.groupby(['Pep', 'Prev', 'Back']).apply(lambda x: ';'.join(list(set(x['Acc'] + '-' + x['StartPos'].astype(str)))))
        group_df = group_df.reset_index()
        group_df.columns = ['Pep', 'Prev', 'Back', 'Acc-PepPos']
        return group_df, not_find_data
    else:
        raise ValueError(f'Return type should be one of the "raw" | "df", now {return_type}')


def batch_add_target_mod(pep_list, mod_type: dict = None, mod_processor=None):
    """
    TODO : This may result of some redundant results (dont know why)
    """
    modpep_list = []
    for pep in rapid_kit.drop_list_duplicates(pep_list):
        modpep_list.extend(add_target_mod(pep, mod_type, mod_processor))
    return rapid_kit.drop_list_duplicates(modpep_list)


def batch_add_target_charge(modpep_list, charge=(2, 3)):
    prec_list = []
    for modpep in rapid_kit.drop_list_duplicates(modpep_list):
        prec_list.extend(add_target_charge(modpep, charge))
    return prec_list


def add_target_mod(pep, mod_type: dict = None, mod_processor=None):
    """
    :param pep: target peptide
    :param mod_type: dict like {'Carbamidomethyl': -1, 'Oxidation': 1}
    where -1 means all possible AAs will be modified with the determined modification
    and an integer means the maximum number of this modification
    :param mod_processor: the mod add processor contains the standard mod rule or customed mod rule
    """
    if mod_type:
        mods = []
        for _mod, _num in mod_type.items():
            _standard_mod = mod_processor.get_standard_mod(_mod)
            possible_aa = mod_processor.query_possible_aa(_standard_mod)
            possible_site = sorted([_.end() for one_aa in possible_aa for _ in re.finditer(one_aa, pep)])
            possible_site_num = len(possible_site)
            if _num > possible_site_num:
                _num = possible_site_num
            if _num == 0 or possible_site_num == 0:
                continue
            elif _num == -1:
                mod = [[(_, _standard_mod) for _ in possible_site]]
            else:
                mod = [[], ]
                for _i in range(1, _num + 1):
                    selected_site = [_ for _ in combinations(possible_site, _i)]
                    mod.extend([[(one_site, _standard_mod) for one_site in each_site_comb] for each_site_comb in selected_site])
            if mods:
                new_mod = [_ + __ for _ in mod for __ in mods]
                mods.extend(new_mod)
            else:
                mods = mod
        if mods:
            mod_pep = [mod_processor.add_mod(pep, one_mod) for one_mod in mods]
        else:
            mod_pep = [pep]
    else:
        mod_pep = [pep]
    return mod_pep


def add_target_charge(modpep, charge=(2, 3)):
    """
    :param modpep: the list of modified peps
    :param charge: a tuple or an integer of targeted charge state
    """
    if isinstance(charge, int):
        charge = (charge,)
    prec_list = []
    for c in charge:
        _prec = rapid_kit.assemble_prec(modpep, c)
        prec_list.append(_prec)
    return prec_list



import os


class GFF3(object):
    """

    https://github.com/The-Sequence-Ontology/Specifications/blob/master/gff3.md

    Generic Feature Format Version 3

    Column 1: "seqid"
    Column 2: "source"
    Column 3: "type"
    Columns 4 & 5: "start" and "end"
    Column 6: "score"
    Column 7: "strand"
    Column 8: "phase"
    Column 9: "attributes"

    Predefined tags in col 9 (attributes)
        ID
        Name
        Alias
        Parent
        Target
        Gap
        Derives_from
        Note
        Dbxref
        Ontology_term
        Is_circular

    Characters with reserved meanings in column 9
        ; semicolon (%3B)
        = equals (%3D)
        & ampersand (%26)
        , comma (%2C)

    Attributes which can have multiple values (split by comma - ,)
        Parent, Alias, Note, Dbxref ,and Ontology_term

    Attribute names are case sensitive.
        Parent != parent

    Lines beginning with '##' are directives (sometimes called pragmas or meta-data) and provide meta-information about the document as a whole.
    Blank lines should be ignored by parsers and lines beginning with a single '#' are used for human-readable comments and can be ignored by parsers.
    End-of-line comments (comments preceded by # at the end of and on the same line as a feature or directive line) are not allowed.

    The Gap Attribute - Eg. Gap=M8 D3 M6 I1 M6
        Code	Operation
        M	match
        I	insert a gap into the reference sequence
        D	insert a gap into the target (delete from reference)
        F	frameshift forward in the reference sequence
        R	frameshift reverse in the reference sequence

    ###
        This directive (three # signs in a row) indicates that all forward references to feature IDs that have been seen to this point have been resolved.
        After seeing this directive, a program that is processing the file serially can close off any open objects that it has created and return them,
        thereby allowing iterative access to the file.

    ##FASTA
        This notation indicates that the annotation portion of the file is at an end and that the remainder of the file contains one or more sequences
        (nucleotide or protein) in FASTA format.
        This allows features and sequences to be bundled together.
        All FASTA sequences included in the file must be included together at the end of the file and may not be interspersed with the features lines.
        Once a ##FASTA section is encountered no other content beyond valid FASTA sequence is allowed.
    """

    def __init__(
            self, gff_path=None, parse_with_init=True,
            sep_symbol=(';',), equal_symbol='=',
            item_sep_symbol=(',',), value_sep_symbol=('&',)
    ):
        self.file_path = gff_path

        if self.file_path:
            self._check_path()

        self.comment_lines = []

        self._have_fasta = False
        self.fasta_lines = None

        if parse_with_init:
            self.parse()

    def _check_path(self):
        if not os.path.exists(self.file_path):
            print(f'The path of input GFF file is not valid: {self.file_path}')
            raise FileNotFoundError

    def set_file_path(self, gff_path):
        self.file_path = gff_path
        self._check_path()

    def parse(self):
        with open(self.file_path, 'r') as f:
            for row in f:
                row = row.strip('\n')
                if row.startswith('#'):
                    self.comment_lines.append(row)

                    if row == '##FASTA':
                        self._have_fasta = True
                        self.fasta_lines = []

                elif row == '###':
                    pass
                else:
                    if self._have_fasta:
                        self.fasta_lines.append(row)
                    else:
                        pass

    def a(self, gff_path):
        gff_dict = dict()
        with open(gff_path, 'r') as f:
            _ = f.readline()
            trans_sites = []
            for r in f:
                if r.startswith('##'):
                    if trans_sites:
                        trans_sites.append(end_site)
                        gff_dict[acc] = trans_sites
                    trans_sites = []
                    split_row = r.split(' ')
                    acc = split_row[1]
                    end_site = int(split_row[3])
                else:
                    split_row = r.split('\t')
                    if split_row[2] == 'Transmembrane':
                        trans_sites.append((int(split_row[3]), int(split_row[4])))
                    else:
                        continue
            trans_sites.append(end_site)
            gff_dict[acc] = trans_sites

    def with_fasta(self):
        if self._have_fasta:
            return True
        else:
            return False



import os
import re

import numpy as np
import pandas as pd

from mskit import calc
from mskit import rapid_kit
from .sn_constant import SNLibraryTitle

__all__ = [
    'filter_inten_neutral_loss_pos',
    'sn_modpep_to_intseq',
    'intseq_to_sn_modpep',
    'get_lib_prec',
    'select_target_run_df',
    'get_library_info',
    'merge_lib',
    'norm_params',
    'get_lib_rt_info',
    'get_lib_fragment_info',
    'write_lib',
]


def filter_inten_neutral_loss_pos(modpep, intens,
                                  exclude_mod_in_pep='[Phospho (STY)]',
                                  exclude_res=('S', 'T'),
                                  exclude_frag_losstype='H3PO4'):
    """
    TODO 这里只是对最小和最大可以出现磷酸根丢失的位点进行 filter，如果需要进一步对应磷酸化修饰数量，应该指定 b/y 之后建一个每个位置的数量 list
    TODO 如果目的不止是为了筛选正确的丢失位点，则最好直接产生一条肽段所有可能的 fragments

    *Suspend*

    :param modpep: This should be a phosphopeptide with general format like _XXXXXS[Phospho (STY)]XXXX_
    :param intens: a intensity dict
    """
    strip_pep, mod_pos, mods = rapid_kit.substring_finder(modpep.replace('_', ''))
    phos_mod_pos = [mod_pos[i] for i, mod in enumerate(mods) if mod == exclude_mod_in_pep]
    st_phosmod_pos = [s for s in phos_mod_pos if strip_pep[s - 1] in exclude_res]
    if st_phosmod_pos:
        positive_min_phos_site = min(st_phosmod_pos)
        negetive_min_phos_site = len(strip_pep) - max(st_phosmod_pos) + 1
    else:
        positive_min_phos_site = len(strip_pep) + 1
        negetive_min_phos_site = len(strip_pep) + 1
    new_frag_dict = dict()
    for frag_name, inten in intens.items():
        frag_type, frag_num, frag_charge, frag_losstype = re.findall(r'([by])(\d+)\+(\d)-(.+)', frag_name)[0]
        if ',' in frag_losstype:
            frag_losstype = frag_losstype.split(',')[1]
        if frag_losstype == exclude_frag_losstype:
            if frag_type == 'b' and frag_num < positive_min_phos_site:
                continue
            elif frag_type == 'y' and frag_num < negetive_min_phos_site:
                continue
            else:
                raise
        new_frag_dict[frag_name] = inten


def filter_inten_dict_with_phosloss(inten):
    pass


def sn_modpep_to_intseq(x):
    x = x.replace('_', '')
    if '[Acetyl (Protein N-term)]' in x:
        x = x.replace('[Acetyl (Protein N-term)]', '')
        x = '*' + x
    else:
        x = '@' + x
    x = x.replace('C[Carbamidomethyl (C)]', 'C')
    x = x.replace('M[Oxidation (M)]', '1')
    x = x.replace('S[Phospho (STY)]', '2')
    x = x.replace('T[Phospho (STY)]', '3')
    x = x.replace('Y[Phospho (STY)]', '4')
    return x


def intseq_to_sn_modpep(x):
    if x[0] == '*':
        x = '[Acetyl (Protein N-term)]' + x[1:]
    elif x[0] == '@':
        x = x[1:]
    else:
        pass
    x = x.replace('C', 'C[Carbamidomethyl (C)]')
    x = x.replace('1', 'M[Oxidation (M)]')
    x = x.replace('2', 'S[Phospho (STY)]')
    x = x.replace('3', 'T[Phospho (STY)]')
    x = x.replace('4', 'Y[Phospho (STY)]')
    x = f'_{x}_'
    return x


def get_lib_prec(lib_df):
    prec = lib_df.apply(lambda x: x['ModifiedPeptide'] + '.' + str(x['PrecursorCharge']), axis=1)
    return prec


def select_target_run_df(df, region_ident, return_col_list=None):
    colname = 'R.Instrument (parsed from filename)'
    return rapid_kit.extract_df_with_col_ident(df, colname, region_ident, return_col_list)


def get_library_info(lib_path):
    lib_df = pd.read_csv(lib_path, sep='\t', low_memory=False)
    protein_groups_num = len(lib_df['ProteinGroups'].drop_duplicates())
    precursor_num = len((lib_df['PrecursorCharge'].astype(str) + lib_df['ModifiedPeptide']).drop_duplicates())
    modpep_num = len(lib_df['ModifiedPeptide'].drop_duplicates())
    return protein_groups_num, precursor_num, modpep_num


def merge_lib(main_lib, accomp_lib, drop_col=('ProteinGroups',)):
    """
    :param main_lib: main library
    :param accomp_lib: accompanying library. Overlapped precursors with main lib in this lib will be deleted
    :param drop_col:
    :return:
    """
    if not isinstance(main_lib, pd.DataFrame):
        main_lib = pd.read_csv(main_lib, sep='\t')
    if not isinstance(accomp_lib, pd.DataFrame):
        accomp_lib = pd.read_csv(accomp_lib, sep='\t')

    main_prec = set(main_lib.apply(lambda x: x['ModifiedPeptide'] + str(x['PrecursorCharge']), axis=1))
    accomp_lib = accomp_lib[~(accomp_lib['ModifiedPeptide'] + accomp_lib['PrecursorCharge'].astype(str)).isin(main_prec)]

    hybrid_lib = main_lib.append(accomp_lib)
    if drop_col:
        hybrid_lib = hybrid_lib.drop(drop_col, axis=1)
    return hybrid_lib


def norm_params(original_df, params_list, norm_func=np.nanmedian, focus_col='EG.ModifiedPeptide'):
    """
    This function is used to get the median value or others depends on the norm func of given params list when one focus is detected in multi reps
    """
    grouped_df = original_df.groupby(focus_col)
    norm_param_df = grouped_df[params_list].transform(norm_func)
    norm_colname = [_ + '_norm' for _ in params_list]
    norm_param_df.columns = norm_colname
    normed_df = pd.concat([original_df, norm_param_df], axis=1)
    return normed_df


def get_lib_rt_info(lib, pep_col='ModifiedPeptide', rt_col='iRT', return_type='dict'):
    rt_df = lib[[pep_col, rt_col]]
    non_redundant_rt_df = rt_df.drop_duplicates(pep_col)
    rt_dict = dict(non_redundant_rt_df.set_index(pep_col)[rt_col])
    if return_type == 'dict':
        return rt_dict
    elif return_type == 'list':
        return [(k, v) for k, v in rt_dict.items()]
    else:
        raise


def get_lib_fragment_info(lib, norm=False):
    """
    Extract fragment intensity info of precursors from library
    :param lib: spectral library path or dataframe
    :param norm:
    :return: A dict has precursors (modpep.charge) as key, and fragment info (a dict has key-value pair like 'b5+1-noloss': 21.1) as value
    """
    if not isinstance(lib, pd.DataFrame):
        if os.path.isfile(lib):
            lib = pd.read_csv(lib, sep='\t')
        else:
            raise
    if 'Precursor' not in lib.columns:
        lib['Precursor'] = lib.apply(lambda x: x['ModifiedPeptide'] + '.' + str(x['PrecursorCharge']), axis=1)

    def _get_group_frag_data(one_group):
        frags = one_group.apply(lambda x: x['FragmentType'] + str(x['FragmentNumber']) + '+' + str(x['FragmentCharge']) + '-' + x['FragmentLossType'], axis=1)
        intens = one_group['RelativeIntensity']
        if norm:
            calc.normalize_intensity(list(intens), norm)
        return dict(zip(frags, intens))

    frag_info = lib.groupby('Precursor').apply(_get_group_frag_data)
    frag_dict = dict(frag_info)
    return frag_dict


def write_lib(output_path, inten_dict: dict, rt_dict: dict, seq2protein_dict=None):
    """
    This function is based on the precursors in inten_dict.
    Seq not in rt_dict will pass
    Seq not in seq2protein_dict will make protein as ''
    Fragment with no loss type will be noloss
    :param output_path:
    :param inten_dict:
    :param rt_dict:
    :param seq2protein_dict:
    :return:
    """
    if seq2protein_dict:
        title_list = SNLibraryTitle.LibraryMainCol
    else:
        title_list = SNLibraryTitle.LibraryMainColPGOut

    with open(output_path, 'w') as _out:
        lib_title = '\t'.join(title_list)
        _out.write(lib_title + '\n')

        for each_prec, frag_dict in inten_dict.items():
            modpep, charge = rapid_kit.split_prec(each_prec, keep_underline=True)
            if modpep in rt_dict:
                irt = rt_dict[modpep]
            else:
                continue
            stripped_pep, mod = rapid_kit.split_mod(modpep, mod_ident='bracket')
            mod = mod.split(' ')[0]
            prec_mz = calc.calc_prec_mz(stripped_pep, charge, mod)
            if seq2protein_dict:
                protein_acc = seq2protein_dict[stripped_pep] if stripped_pep in seq2protein_dict else ''
                protein_acc = protein_acc if isinstance(protein_acc, str) else ';'.join(set(protein_acc))
            else:
                protein_acc = ''
            for frag_name, frag_inten in frag_dict.items():
                frag_type, frag_num, frag_charge, frag_loss = rapid_kit.split_fragment_name(frag_name)
                frag_loss = frag_loss if frag_loss else 'noloss'
                fragment_mz = calc.calc_fragment_mz(stripped_pep, frag_type, frag_num, frag_charge, mod=mod)
                one_row_list = [charge, modpep, stripped_pep, irt,
                                modpep, prec_mz, frag_loss, frag_num,
                                frag_type, frag_charge, fragment_mz, frag_inten]
                one_row_list = list(map(str, one_row_list))
                if seq2protein_dict:
                    one_row_list.append(protein_acc)
                lib_line = "\t".join(one_row_list)
                _out.write(lib_line + '\n')



import os
import re
from collections import defaultdict

import pandas as pd

from mskit import rapid_kit
from mskit.post_analysis.spectronaut import SpectronautLibrary
from ._pdeep_constant import BasicpDeepInfo
from ._pdeep_constant import MOD


def intprec_to_pdeep_test(intprec_list):
    """
    从 intprec 转换为 pDeep2 的 test input
    intprec: 如 DKEAIQA4SESLMTSAPK.2
    pDeep2 test input 格式为
        peptide modification    charge
        FRTPSFLK    3,Phospho[T];5,Phospho[S];  2
        ...
    """
    title = ['peptide', 'modification', 'charge']
    pdeep_test_data_list = []
    for each_intprec in intprec_list:
        intprec_result = intprec_to_pdeep(each_intprec)
        if intprec_result is not None:
            stripped_pep, mod_info, charge = intprec_result
        else:
            continue
        pdeep_test_data_list.append([stripped_pep, mod_info, charge])
    pdeep_test_df = pd.DataFrame(pdeep_test_data_list, columns=title)
    return pdeep_test_df


def intprec_to_pdeep(intprec: str):
    int_to_pdeep2_mod = {
        'C': 'Carbamidomethyl[C]',
        '1': 'Oxidation[M]',
        '2': 'Phospho[S]',
        '3': 'Phospho[T]',
        '4': 'Phospho[Y]',
    }
    intseq, charge = intprec.split('.')
    if intseq.startswith('@'):
        intseq = intseq[1:]
    elif intseq.startswith('*'):
        return None
    else:
        pass
    stripped_pep = intseq.replace('1', 'M').replace('2', 'S').replace('3', 'T').replace('4', 'Y')
    mod_info = ''
    for _ in re.finditer('[C1234]', intseq):
        site = _.end()
        mod_char = _.group()
        mod = int_to_pdeep2_mod[mod_char]
        mod_info += f'{site},{mod};'
    return stripped_pep, mod_info, charge


def mod_extraction_for_pdeep(mod_pep):
    """
    The
    """
    mod_pep = mod_pep.replace('_', '')
    if '[' not in mod_pep:
        return ''
    else:
        modinfo = ''
        mod_start = [left_bracket.start() for left_bracket in re.finditer('\[', mod_pep)]
        mod_end = [right_bracket.start() for right_bracket in re.finditer(']', mod_pep)]
        mod_len = 0
        for mod_site in zip(mod_start, mod_end):
            if mod_site[0] == 0:  # or mod_site[1] == len(mod_pep) - 1:
                return 'Unsupport'
            else:
                mod_residu = mod_pep[mod_site[0] - 1]
                mod_type = mod_pep[mod_site[0] + 1: mod_site[1]].replace(' ', '')
                mod_type = re.sub(r'\(.+?\)', f'[{mod_residu}]', mod_type)
                modinfo += '{mod_site},{mod_type};'.format(mod_site=mod_site[0] - mod_len, mod_type=mod_type)
                mod_len += (mod_site[1] - mod_site[0] + 1)
        return modinfo


def inten_dict_to_plabel(inten_dict: dict):
    """
    :param inten_dict: The input dict should have the k[v] pairs as 'Prec': {'Frag_1': Inten_1, 'Frag_2': Inten_2, ...}
    """
    plabel_rows = []
    for prec, ion_inten_dict in inten_dict.items():
        intprec_trans = intprec_to_pdeep(prec)
        if intprec_trans is None:
            continue
        stripped_pep, mod_info, charge = intprec_trans
        spec = f'Unknown.{charge}.0.0'
        plabel_ion_str = plabel_one_ion_row(ion_inten_dict, return_type='str')
        plabel_rows.append(f'{spec}\t{stripped_pep}\t{mod_info}\t{plabel_ion_str}')
    return plabel_rows


def write_plabel_with_inten_dict(inten_dict: dict, output_path: str):
    plabel_rows = inten_dict_to_plabel(inten_dict)
    with open(output_path, 'w') as f:
        f.write('spec\tpeptide\tmodinfo\tb\tb-NH3\tb-H2O\tb-ModLoss\ty\ty-NH3\ty-H2O\ty-ModLoss\n')
        for row in plabel_rows:
            f.write(row + '\n')


def plabel_to_pred_input(plabel_path):
    plabel_df = pd.read_csv(plabel_path, sep='\t', low_memory=False)
    plabel_df['charge'] = plabel_df['spec'].apply(lambda x: x.split('.')[-3])
    plabel_df = plabel_df[['peptide', 'modinfo', 'charge']]
    plabel_df.columns = ['peptide', 'modification', 'charge']
    return plabel_df


def plabel_one_ion_row(ion_inten_dict: dict,
                       ion_type=('b', 'b-NH3', 'b-H2O', 'b-ModLoss', 'y', 'y-NH3', 'y-H2O', 'y-ModLoss'),
                       return_type='str'):
    ion_dict = defaultdict(list)
    ion_dict.fromkeys(ion_type)
    loss_trans = {'1,H3PO4': 'ModLoss',
                  '1,H2O': 'H2O',
                  '1,NH3': 'NH3'}
    for frag, inten in ion_inten_dict.items():
        frag_type, frag_num, frag_charge, frag_loss = re.findall(r'([abcxyz])(\d+)\+(\d)-(.+)', frag)[0]
        if frag_loss == 'Noloss':
            ion_name = f'{frag_type}'
        elif frag_loss in ['1,H2O', '1,NH3', '1,H3PO4']:
            ion_name = f'{frag_type}-{loss_trans[frag_loss]}'
        else:
            continue
        ion_dict[ion_name].append((f'{frag_type}{frag_num}{ion_name[1:]}+{frag_charge},{inten};',
                                   int(frag_num),
                                   int(frag_charge)))
    if return_type == 'dict':
        return ion_dict
    elif return_type == 'str':
        ion_info = []
        for each_ion_type in ion_type:
            if each_ion_type[0] in ['a', 'b', 'c']:
                sorted_ions = sorted(ion_dict[each_ion_type], key=lambda x: (x[2], x[1]), reverse=False)
            elif each_ion_type[0] in ['x', 'y', 'z']:
                sorted_ions = sorted(ion_dict[each_ion_type], key=lambda x: (-x[2], x[1]), reverse=True)
            else:
                raise
            ions = [_[0] for _ in sorted_ions]
            ion_info.append(''.join(ions))
        return '\t'.join(ion_info)


def plabel_ion_info(one_psm_df, return_type):
    ion_info = {'b': '', 'b-NH3': '', 'b-H2O': '', 'b-ModLoss': '', 'y': [], 'y-NH3': [], 'y-H2O': [], 'y-ModLoss': []}
    for row_index, each_row in one_psm_df.iterrows():
        fragment_type = each_row['FragmentType']
        fragment_num = each_row['FragmentNumber']
        fragment_charge = each_row['FragmentCharge']
        fragment_relative_intensity = each_row['RelativeIntensity']
        fragment_losstype = each_row['FragmentLossType']
        if fragment_type == 'b':
            if fragment_losstype == 'noloss':
                ion_info['b'] += 'b{num}+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge, relative_intensity=fragment_relative_intensity)
            elif fragment_losstype == 'NH3':
                ion_info['b-NH3'] += 'b{num}-NH3+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge, relative_intensity=fragment_relative_intensity)
            elif fragment_losstype == 'H2O':
                ion_info['b-H2O'] += 'b{num}-H2O+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge, relative_intensity=fragment_relative_intensity)
            elif fragment_losstype == 'H3PO4':
                ion_info['b-ModLoss'] += 'b{num}-ModLoss+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                                relative_intensity=fragment_relative_intensity)
            else:
                continue
        elif fragment_type == 'y':
            if fragment_losstype == 'noloss':
                ion_info['y'].append('y{num}+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge, relative_intensity=fragment_relative_intensity))
            elif fragment_losstype == 'NH3':
                ion_info['y-NH3'].append(
                    'y{num}-NH3+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge, relative_intensity=fragment_relative_intensity))
            elif fragment_losstype == 'H2O':
                ion_info['y-H2O'].append(
                    'y{num}-H2O+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge, relative_intensity=fragment_relative_intensity))
            elif fragment_losstype == 'H3PO4':
                ion_info['y-ModLoss'].append(
                    'y{num}-ModLoss+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge, relative_intensity=fragment_relative_intensity))
            else:
                continue
    if return_type == 'dict':
        return ion_info
    elif return_type == 'str':
        str_ion_info = ''
        b_ion_order = ['b', 'b-NH3', 'b-H2O', 'b-ModLoss']
        #   ion_order = ['b', 'y']
        for ion_losstype in b_ion_order:
            str_ion_info += ion_info[ion_losstype]
            str_ion_info += '\t'

        y_ion_order = ['y', 'y-NH3', 'y-H2O', 'y-ModLoss']
        for ion_losstype in y_ion_order:
            str_ion_info += ''.join(ion_info[ion_losstype][::-1])
            if ion_losstype != 'y-ModLoss':
                str_ion_info += '\t'
            #   if ion_losstype != 'y':
            #       str_ion_info += '\t'

        return str_ion_info


def sn_lib_to_plabel(lib, plabel_output):
    if isinstance(lib, pd.DataFrame):
        lib_df = lib
    else:
        if os.path.exists(lib):
            lib_df = pd.read_csv(lib, sep='\t', low_memory=False)
        else:
            raise FileNotFoundError
    lib_df['Prec'] = lib_df['ModifiedPeptide'] + '.' + lib_df['PrecursorCharge'].astype(str)
    with open(plabel_output, 'w') as plabel_handle:
        plabel_handle.write('spec\tpeptide\tmodinfo\tb\tb-NH3\tb-H2O\tb-ModLoss\ty\ty-NH3\ty-H2O\ty-ModLoss\n')
        #   handle_plabel.write('spec\tpeptide\tmodinfo\tb\ty\n')
        for psm_index, (each_prec, each_psm_df) in enumerate(lib_df.groupby('Prec')):
            first_row = each_psm_df.iloc[0]
            spec = '{title}.{charge}.0.0'.format(title=first_row['ReferenceRun'], charge=first_row['PrecursorCharge'])
            #   spec = '{charge}.0.0'.format(charge=first_fragment[1])
            stripped_pep = first_row['StrippedPeptide']
            mod_pep = first_row['ModifiedPeptide']
            modinfo = mod_extraction_for_pdeep(mod_pep)
            if modinfo == 'Unsupport':
                continue
            ion_info = plabel_ion_info(each_psm_df, 'str')
            plabel_handle.write('{spec}\t{pep}\t{mod}\t{ioninfo}\n'.format(
                spec=spec, pep=stripped_pep, mod=modinfo, ioninfo=ion_info))


def sn_lib_to_pdeep_test(test_lib, test_set_output):
    if isinstance(test_lib, pd.DataFrame):
        lib_df = test_lib
    else:
        if os.path.exists(test_lib):
            lib_df = pd.read_csv(test_lib, sep='\t', low_memory=False)
        else:
            raise FileNotFoundError
    lib_df['Prec'] = lib_df['ModifiedPeptide'] + '.' + lib_df['PrecursorCharge'].astype(str)
    lib_df = lib_df.drop_duplicates('Prec')
    with open(test_set_output, 'w') as test_handle:
        test_handle.write('peptide\tmodification\tcharge\n')

        for row_index, each_lib_row in lib_df.iterrows():
            mod_pep = each_lib_row['ModifiedPeptide']
            charge = str(each_lib_row['PrecursorCharge'])
            stripped_pep = each_lib_row['StrippedPeptide']
            mod = mod_extraction_for_pdeep(mod_pep)
            if mod == 'Unsupport':
                continue
            test_handle.write('{}\t{}\t{}\n'.format(stripped_pep, mod, charge))


def extract_pdeep_mod(mod_pep, mod_ident='bracket', mod_trans=True):
    """
    input: '_C[Carbamidomethyl (C)]DM[Oxidation (M)]EDER_'
    output: 'CDMEDER', '1,Carbamidomethyl[C];3,Oxidation[M];'
    """
    stripped_pep, mod = rapid_kit.split_mod(modpep=mod_pep, mod_ident=mod_ident)
    if mod_trans:
        mod = trans_sn_mod(mod)
    return stripped_pep, mod


def trans_sn_mod(mod):
    for sn_mod, pdeep_mod in MOD.items():
        mod = mod.replace(sn_mod, pdeep_mod)
        if '(' not in mod:
            break
    if '(' in mod:
        return None
    return mod


def restore_pdeep_mod_site(stripped_pep, mod_content, mod_processor):
    """
    This will restore the modification to stripped peptide.
    EXAMPLE: restore_pdeep_mod_site('MPALAIMGLSLAAFLELGMGASLCLSQQFK', '24,Carbamidomethyl[C];')
    -> 'MPALAIMGLSLAAFLELGMGASLC[Carbamidomethyl (C)]LSQQFK'
    """
    return rapid_kit.add_mod(stripped_pep, mod_content, mod_processor)


def pdeep_input(output_path, prec_list):
    with open(output_path, 'w') as out_file:
        pred_title = ['peptide', 'modification', 'charge']
        out_file.write('\t'.join(pred_title) + '\n')
        for _prec in prec_list:
            modpep, charge = rapid_kit.split_prec(_prec)
            strip_pep, mod = extract_pdeep_mod(modpep)
            out_file.write(f'{strip_pep}\t{mod}\t{charge}\n')


def pdeep_trainset(output_path, prec_inten_dict):
    with open(output_path, 'w') as out_file:
        plabel_title_list = BasicpDeepInfo.pDeepTrainsetTitle
        plabel_title = '\t'.join(plabel_title_list)
        out_file.write(plabel_title + '\n')
        for _prec, inten_dict in prec_inten_dict.items():
            plabel_row_dict = plabel_one_row_dict(_prec, inten_dict)
            if not plabel_row_dict:
                continue
            one_row_list = [plabel_row_dict[_] for _ in plabel_title_list]
            out_file.write('\t'.join(one_row_list) + '\n')


def plabel_one_row_dict(prec, inten_dict: dict):
    plabel_row_dict = defaultdict(str)
    modpep, charge = rapid_kit.split_prec(prec)
    strip_pep, mod = extract_pdeep_mod(modpep, mod_ident='bracket', mod_trans=True)
    if not mod:
        return None
    plabel_row_dict['spec'] = f'{charge}.0.0'
    plabel_row_dict['peptide'] = strip_pep
    plabel_row_dict['modinfo'] = mod
    for frag, inten in inten_dict.items():
        frag_type, frag_num, frag_charge, frag_loss = rapid_kit.split_fragment_name(frag)
        if frag_loss == 'noloss':
            plabel_type = frag_type
            plabel_frag = f'{frag_type}{frag_num}+{frag_charge}'
        elif frag_loss == 'NH3' or frag_loss == 'H2O':
            plabel_type = f'{frag_type}-{frag_loss}'
            plabel_frag = f'{frag_type}{frag_num}-{frag_loss}+{frag_charge}'
        else:
            plabel_type = f'{frag_type}-ModLoss'
            plabel_frag = f'{frag_type}{frag_num}-ModLoss+{frag_charge}'
        plabel_row_dict[plabel_type] += f'{plabel_frag},{inten};'
    return plabel_row_dict


def read_pdeep_result(pdeep_result, modloss_name='H3PO4',
                      require_mz=True, min_inten_ratio=0.01, min_frag_num=3,
                      exclude_frag_num=(1, 2), exclude_modloss=False):
    mod_dict = {'Carbamidomethyl[C]': '[Carbamidomethyl (C)]',
                'Oxidation[M]': '[Oxidation (M)]',
                'Phospho[S]': '[Phospho (STY)]',
                'Phospho[T]': '[Phospho (STY)]',
                'Phospho[Y]': '[Phospho (STY)]',
                }
    with open(os.path.abspath(pdeep_result), 'r') as pdeep_handle:
        predicted_fragment_data = dict()
        for each_line in pdeep_handle:
            each_line = each_line.strip('\n')
            if each_line == 'BEGIN IONS':
                fragment_dict = dict()
            elif each_line == 'END IONS':
                if len(fragment_dict) >= min_frag_num:
                    predicted_fragment_data[prec] = fragment_dict
                else:
                    pass
            else:
                if each_line.startswith('TITLE'):
                    split_pep_title = each_line.replace('TITLE=', '').split('|')
                    stripped_pep = split_pep_title[0]
                    mod = split_pep_title[1].strip(';')
                    charge = split_pep_title[2]

                    if not mod:
                        prec = '_{}_.{}'.format(stripped_pep, charge)
                    else:
                        mod_pep = ''
                        previous_mod_site = 0
                        for each_mod in mod.split(';'):
                            each_mod_info = each_mod.split(',')
                            mod_site = int(each_mod_info[0])
                            mod_type = mod_dict[each_mod_info[1]]
                            mod_pep += stripped_pep[previous_mod_site: mod_site] + mod_type
                            previous_mod_site = mod_site
                        mod_pep += stripped_pep[previous_mod_site:]
                        prec = '_{}_.{}'.format(mod_pep, charge)
                elif each_line[0].isdigit():
                    split_frag_inten_line = each_line.split(' ')
                    frag_inten = round(float(split_frag_inten_line[1]), 5) * 100
                    if frag_inten < min_inten_ratio:
                        continue
                    frag_mz = split_frag_inten_line[0]
                    if float(frag_mz) < 10:
                        continue
                    frag_name = split_frag_inten_line[2]

                    frag_type, frag_num, loss_type, frag_c = re.findall('([by])(\d+)-?(.+)?\+(\d)', frag_name)[0]
                    if int(frag_num) in exclude_frag_num:
                        continue
                    if exclude_modloss and loss_type == 'ModLoss':
                        continue
                    new_frag_name = f'{frag_type}{frag_num}+{frag_c}'
                    if not loss_type:
                        new_frag_name += '-noloss'
                    else:
                        new_frag_name += f'-{loss_type}' if loss_type != 'ModLoss' else f'-{modloss_name}'
                    if require_mz:
                        fragment_dict[new_frag_name] = (frag_mz, frag_inten)
                    else:
                        fragment_dict[new_frag_name] = frag_inten
                else:
                    continue
    return predicted_fragment_data


def trans_pdeep2_result_to_df(result: dict, frag_trans=None, pep_trans=None, pep_trans_col='IntPep') -> pd.DataFrame:
    df_rows = []
    for prec, inten_dict in result.items():
        if frag_trans is not None:
            inten_dict = {frag_trans[frag]: inten for frag, inten in inten_dict.items()}
        one_row = [prec, inten_dict]
        if pep_trans is not None:
            modpep, charge = prec.split('.')
            transed_pep = pep_trans(modpep)
            one_row.append(transed_pep)


def read_inten_from_plabel(_plabel_file):
    ion_type_list = ['b', 'b-NH3', 'b-H2O', 'b-ModLoss', 'y', 'y-NH3', 'y-H2O', 'y-ModLoss']
    _p_df = pd.read_csv(_plabel_file, sep='\t')
    _p_df = _p_df.fillna('')
    _p_df['prec'] = _p_df.apply(lambda x: '|'.join([x['peptide'], x['modinfo'], x['spec'].split('.')[-3]]), axis=1)

    _p_inten_dict = dict()

    def _merge_plabel_inten(x):
        _one_prec = x['prec']
        _one_inten_info = ''.join(x[ion_type_list].tolist()).split(';')[:-1]
        _p_inten_dict[_one_prec] = dict([(_o_f.split(',')[0], float(_o_f.split(',')[1])) for _o_f in _one_inten_info])

    _p_df.progress_apply(_merge_plabel_inten, axis=1)
    return _p_inten_dict


class pDeepSpectronaut(SpectronautLibrary):
    def __init__(self, spectronaut_version=12):
        super(pDeepSpectronaut, self).__init__(spectronaut_version)
        self.plabel_title_list = BasicpDeepInfo.pDeepTrainsetTitle

    def prec_ion_info(self, one_psm_df: pd.DataFrame, spectronaut_run_name=True):
        """
        For pDeep trainset preparation.
        This will receive get_one_prefix_result dataframe of one psm block and assemble get_one_prefix_result pd.series as one row of the plabel dataframe.
        :param one_psm_df: This must contain columns after ['PrecursorCharge', 'StrippedPeptide', 'ModifiedPeptide',
        'FragmentType', 'FragmentNumber', 'FragmentCharge', 'RelativeIntensity', 'FragmentLossType']
        :param spectronaut_run_name: This can be choose as True or False and dont affect the result. This can make the plabel file have much information
        :return: A series as one plabel dataframe row
        """
        first_row = one_psm_df.iloc[0]
        prec_charge = first_row['PrecursorCharge']
        if spectronaut_run_name:
            run_title = first_row['ReferenceRun']
            spec = '{title}.{charge}.0.0'.format(title=run_title, charge=prec_charge)
        else:
            spec = '{charge}.0.0'.format(charge=prec_charge)

        stripped_pep = first_row['StrippedPeptide']
        mod_pep = first_row['ModifiedPeptide']
        stripped_pep, modinfo = extract_pdeep_mod(mod_pep)
        if modinfo == 'Unsupport':
            return 'Unsupport'
        current_prec_info = pd.Series(data=[spec, stripped_pep, modinfo] + [''] * 8, index=self.plabel_title_list)

        for row_index in one_psm_df.index:
            line_series = one_psm_df.loc[row_index]

            fragment_type = line_series['FragmentType']
            fragment_num = line_series['FragmentNumber']
            fragment_charge = line_series['FragmentCharge']
            fragment_relative_intensity = line_series['RelativeIntensity']
            fragment_losstype = line_series['FragmentLossType']
            if fragment_type == 'b':
                if fragment_losstype == 'noloss':
                    current_prec_info['b'] += 'b{num}+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                             relative_intensity=fragment_relative_intensity)
                elif fragment_losstype == 'NH3':
                    current_prec_info['b-NH3'] += 'b{num}-NH3+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                                     relative_intensity=fragment_relative_intensity)
                elif fragment_losstype == 'H2O':
                    current_prec_info['b-H2O'] += 'b{num}-H2O+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                                     relative_intensity=fragment_relative_intensity)
                else:
                    current_prec_info['b-ModLoss'] += 'b{num}-ModLoss+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                                             relative_intensity=fragment_relative_intensity)
            elif fragment_type == 'y':
                if fragment_losstype == 'noloss':
                    current_prec_info['y'] += 'y{num}+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                             relative_intensity=fragment_relative_intensity)
                elif fragment_losstype == 'NH3':
                    current_prec_info['y-NH3'] += 'y{num}-NH3+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                                     relative_intensity=fragment_relative_intensity)
                elif fragment_losstype == 'H2O':
                    current_prec_info['y-H2O'] += 'y{num}-H2O+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                                     relative_intensity=fragment_relative_intensity)
                else:
                    current_prec_info['y-ModLoss'] += 'y{num}-ModLoss+{charge},{relative_intensity};'.format(num=fragment_num, charge=fragment_charge,
                                                                                                             relative_intensity=fragment_relative_intensity)
        return current_prec_info

    def plabel_trainset(self, output_path, spectronaut_run_name=True):
        """
        Write get_one_prefix_result pDeep trainset file by calling function prec_ion_info to process the library dataframe
        """
        trainset_df = pd.DataFrame(columns=self.plabel_title_list)
        for each_psm_index in self.get_psm_block_index(self._lib_df):
            current_prec_info = self.prec_ion_info(self._lib_df.loc[each_psm_index[0]: each_psm_index[1]], spectronaut_run_name)
            if not isinstance(current_prec_info, pd.DataFrame):
                continue
            trainset_df = trainset_df.append(current_prec_info, ignore_index=True)
        trainset_df.to_csv(output_path, sep='\t', index=False)


def extract_bracket(str_with_bracket):
    bracket_start = [left_bracket.start() for left_bracket in re.finditer('\(', str_with_bracket)]
    bracket_end = [right_bracket.start() for right_bracket in re.finditer('\)', str_with_bracket)]
    return bracket_start, bracket_end


mod_dict = {'M(ox)': 'Oxidation[M]',
            'Y(ph)': "Phospho[Y]",
            'S(ph)': "Phospho[S]",
            'T(ph)': "Phospho[T]",
            }


def _plabel_from_mq(x):
    def pdeep_mod_extraction(mod_pep):
        mod_pep = mod_pep.replace('_', '')
        modinfo = ''
        mod_start, mod_end = extract_bracket(mod_pep)
        mod_len = 0
        for mod_site in zip(mod_start, mod_end):
            mod_type = mod_pep[mod_site[0] - 1: mod_site[1] + 1].replace(' ', '')
            mod_type = mod_dict[mod_type]
            modinfo += '{mod_site},{mod_type};'.format(mod_site=mod_site[0] - mod_len, mod_type=mod_type)
            mod_len += (mod_site[1] - mod_site[0] + 1)
        return modinfo

    ion_type_list = ['b', 'b-NH3', 'b-H2O', 'b-ModLoss', 'y', 'y-NH3', 'y-H2O', 'y-ModLoss']
    plabel_title = ['spec', 'peptide', 'modinfo', *ion_type_list]

    spec_name = '{}.{}.{}.{}.0.dta'.format(x['Raw file'], x['Scan number'], x['Scan number'], x['Charge'])
    pep = x['Sequence']
    mod_pep = x['Modified sequence']
    mod_info = pdeep_mod_extraction(mod_pep)

    ions = x['Matches']
    intens = x['Intensities']
    inten_dict = dict(zip(ion_type_list, [''] * 8))

    ion_intens_list = list(zip(ions.split(';'), intens.split(';')))
    b_ion_info = [_ for _ in ion_intens_list if _[0].startswith('b')]
    y_ion_info = [_ for _ in ion_intens_list if _[0].startswith('y')]

    for diff_ion_info in [b_ion_info, y_ion_info]:
        current_num = 0
        _mod_start = False
        _second_mod_start = False
        for ion, inten in diff_ion_info:

            if '*' in ion:
                if not _mod_start:
                    current_num = 0
                    _mod_start = True
            if '-' in ion:
                if _mod_start:
                    continue

            ion_type, ion_num = re.findall('([by])(\d+)', ion)[0]
            ion_num = int(ion_num)

            re_charge = re.findall('\((\d)\+\)', ion)
            if re_charge:
                ion_charge = re_charge[0]
            else:
                ion_charge = '1'

            if ion_num <= current_num and '*' in ion:
                _second_mod_start = True
                continue
            if '*' in ion and _second_mod_start:
                continue
            current_num = ion_num

            tag = ion_type
            if '*' in ion:
                tag += '-ModLoss'
            elif '-' in ion:
                tag += '-{}'.format(re.findall('-(.+)', ion)[0])

            inten_dict[tag] += '{}{}{}+{},{};'.format(ion_type,
                                                      ion_num,
                                                      '-' + tag.split('-')[1] if '-' in tag else '',
                                                      ion_charge,
                                                      inten
                                                      )

    one_psm_data = [spec_name, pep, mod_info, *[inten_dict[_] for _ in ion_type_list]]
    return one_psm_data


"""  NOTICE This one is for MQ > 1.6, in which the modifications added in the peptide sequence was set as Phospho (STY) but not (ph) in 1.5

def extract_bracket(str_with_bracket):
    bracket_start = [left_bracket.start() for left_bracket in re.finditer('\(', str_with_bracket)][::2]
    bracket_end = [right_bracket.start() for right_bracket in re.finditer('\)', str_with_bracket)][1::2]
    return bracket_start, bracket_end

mod_dict2 = {'M(Oxidation (M))': 'Oxidation[M]',
            'Y(Phospho (STY))' : "Phospho[Y]",
            'S(Phospho (STY))' : "Phospho[S]",
            'T(Phospho (STY))' : "Phospho[T]",}

def pdeep_mod_extraction(mod_pep):
    mod_pep = mod_pep.replace('_', '')
    modinfo = ''
    mod_start, mod_end = extract_bracket(mod_pep)
    mod_len = 0
    for mod_site in zip(mod_start, mod_end):
        mod_type = mod_pep[mod_site[0] - 1: mod_site[1] + 1]# .replace(' ', '')
        mod_type = mod_dict2[mod_type]
        modinfo += '{mod_site},{mod_type};'.format(mod_site=mod_site[0] - mod_len, mod_type=mod_type)
        mod_len += (mod_site[1] - mod_site[0] + 1)
    return modinfo
"""



"""
Xwalk (version 0.6)
-------------------

INTRODUCTION
------------
Chemical cross-linking of proteins or protein complexes and the mass
spectrometry based localization of the cross-linked amino acids is a powerful
method for generating distance information on the substrate topology. Here we
introduce the algorithm Xwalk for predicting and validating these cross-links on
existing protein structures. Xwalk calculates and displays non-linear distances
between chemically cross-linked amino acids on protein surfaces, while mimicking
the flexibility and non-linearity of cross-linker molecules. It returns a
"Solvent Accessible Surface" (SAS) distance, which corresponds to the length of
the shortest path between two amino acids, where the path leads through solvent
occupied space without penetrating the protein surface.


TEST
-----
Test examples to execute Xwalk can be found in the test subdirectory.


CLASSPATH
---------
You might want to consider adding the bin directory into the CLASSPATH
environment of your SHELL, which avoids the usage of the -cp flag when executing
Xwalk.


COMMANDLINE PARAMETERS
----------------------
A list of all commandline parameters can be retrieved by typing -help before
a Xwalk execution.


OUTPUT
------
Distance information will be printed out to the STDOUT channel or via -out to a
local file in the following tab delimeted format:
...
1   1brs.pdb    LYS-1-D-CB  LYS-2-D-CB  1   5.9 6.6 0.0449  0.0685  KKAVINGEQIR-KAVINGEQIR
...
1st column: index
2nd column: filename
3rd column: PDB information of the 1st amino acid in the format:
            PDBThreeLetterCode-ResidueId-ChainId-AtomName
4th column: PDB information of the 2nd amino acid in the format:
            PDBThreeLetterCode-ResidueId-ChainId-AtomName
5th column: Sequence distance within the PDB file as calculated from the residue
            identifiers of the 1st and 2nd amino acid.
6th column: Euclidean distance between 1st and 2nd amino acid.
7th column: SAS distance between  1st and 2nd amino acid.
8th column: Probability for observing the Euclidean distance in a XL experiment
            using DSS or BS3.
9th column: Probability for observing the SAS distance in a XL experiment using
            DSS or BS3.
10th column: shortest tryptic peptide sequence for both amino acids.

Setting -pymol -out xxx.pml on the commandline will write a PyMOL script into
the file xxx.pml, which can be load into the molecular viewer PyMOL to visualise
the SAS distance paths (see NOTES).


ALGORITHM
---------
The SAS distance is calculated using a grid and the breadth-first search algorithm
to search within the grid for the shortest path between two points on the protein
surface using following algorithm:
1.)    Read in input data
    a. xyz.pdb, spatial coordinates of a protein or protein complex in PDB
       format.
    b. maxdist: maximum distance of the path (i.e. the length of the
       cross-linker + AA side chain length)
    c. listXL, a list of experimentally determined cross-linked lysine residues.
2.)    Remove all non-protein atoms in xyz.pdb and assign protein atoms a van der
    Waals radius sum of SURFNET atom radii + solvent radius
3.)    Select a random lysine pair AAab from listXL,
4.)    Check Euclidean distance (Euc) of AAab. Continue, if Euc > maxdist,
    disregard otherwise and go back to 3.)
5.)    Generate a grid of size maxdist and grid spacing 1 Angstroem centered at AAa
6.)    Set Integer.MAX_VALUE as distance for all grid cells and label grid cells as
    residing in the
    a. protein
    b. solvent
    c. boundary between protein and solvent
7.)    Label grid cells residing in AAab as solvent
8.)    Set distance dist = 0.0 for central grid cell of AAa and store grid cell in
    the active list listactive
9.)    Start breadth-first search. Iterate through listactive
    a. Check that grid cell i is labeled as solvent
    b. Find all immediate neighbors listneighbour
    c. Iterate through listneighbour
        i. Check that grid cell j is labeled as solvent
        ii. Compute new distance for grid cell j as the sum of the distance in
            grid cell i and the Euclidean distance between grid cell i and j
        iii. If distance sum in 9.c.ii is smaller than the current distance in
             grid cell j, store the distance sum as new distance for grid cell j
             and add grid cell j to the new active list listnew_active,
10.) Go back to step 9.) with listactive = listnew_active



NOTES
-----
- As the SAS distance is based on a grid calculation, the default heap size of
  the Java VM with 64MB is likely to be too small. You can increase the heap
  size with "java -Xmx512m"
- You can obtain PyMOL for free at the webpage: http://pymol.org/
- Beware that in order for PyMOL to recognize the script file, the file must
  have the name ending .pml
- You can load the script directly at the startup of PyMOL, i.e. with the
  command pymol 1brs.pml.


CONTACT
-------
abdullah@imsb.biol.ethz.ch


LICENCE
-------
Xwalk executable and libraries are available under the Creative Commons
Attribution-NonCommercial-ShareAlike 3.0 Unported License via the Xwalk website.

Anyone is free:
     to copy, modify, distribute the software;

Under the following conditions:
    - the original authors must be given credit by citing the original Xwalk paper:
        Kahraman A., Malmström L., Aebersold R. (2011). Xwalk: Computing and
        Visualizing Distances in Cross-linking Experiments. Bioinformatics,
        doi:10.1093/bioinformatics/btr348.
    - the software or derivate works must not be used for commercial purposes
    - derivate works must be licenced under the same or similar licence

Any of the above conditions can be waived if the authors give permission.


Xwalk on GitHub
---------------
https://github.com/abxka/Xwalk


COMMAND EXAMPLES
----------------
    # Help with information on the available parameter
java -cp ../bin/ Xwalk -help

    # Example executions on the barnase-barstar complex 1brs.pdb.
    # Calculate the Euclidean distance (-euc)
    # between the closest atoms of lysine residue (-aa1 lys -aa2 lys)
    # that have a maximum distance of 30 Angstroem.
java -cp ../bin/ Xwalk -infile 1brs.pdb -aa1 lys -aa2 lys -euc

    # Calculate the Euclidean distance (-euc)
    # between the closest atoms of aspartic (-aa1 asp) and glutamic (-aa2 glu) acid residues
    # that have a maximum distance of 30 Angstroem.
java -cp ../bin/ Xwalk -infile 1brs.pdb -aa1 asp -aa2 glu -euc -max 21.4

    # Calculate SASD
    # between beta-carbon atoms (-a1 cb -a2 cb)
    # of lysine residues (-aa1 lys -aa2 lys).
    # that have a maximum distance of 21.4 Angstroem (-max 21.4).
    # and remove prior to calculation all side chains (-bb)
java -Xmx256m -cp ../bin/ Xwalk -infile 1brs.pdb -aa1 lys -aa2 lys -a1 cb -a2 cb -max 21.4 -bb

    # Calculate SASD of intermolecular cross-links (-inter)
    # between the epsilon-nitrogen atoms (-a1 nz -a2 nz)
    # of lysine residues (-aa1 lys -aa2 lys).
    # that have a maximum distance of 11.4 Angstroem (-max 11.4).
    # and remove prior to calculation all side chains (-bb)
    # and output PyMOL script to visualize the SASD path (-pymol -out 1brs.pml)
java -Xmx256m -cp ../bin/ Xwalk -infile 1brs.pdb -aa1 lys -aa2 lys -a1 nz -a2 nz -max 21.4 -inter -pymol -out 1brs.pml


Xwalk HELP MESSAGE
------------------
EXAMPLARY command for program execution:
Xwalk -in 1brs.pdb -aa1 ARG -aa2 lys -a1 CB -a2 CB -bb -max 21

ABOUT
Version 0.6
Xwalk calculates and outputs distances in Angstroem for potential cross-links
between -aa1 type amino acids and -aa2 type amino acids in the PDB file -in.

IMPORTANT
If large protein complexes are processed, the Java heap size might need to be
increased from the default 64MB to 256MB, with the Java parameter -Xmx256m

OUTPUT FORMAT:
IndexNo InfileName      Atom1info       Atom2info       DistanceInPDBsequence   EuclideanDistance       SolventAccessibleSurfaceDistance        EucProbability  SASDprobability PeptidePairSequences

where the Solvent Accessible Surface (SAS) distance corresponds to a number code, when a Distance file (-dist) is provided:
        >= 0: SAS distance (when -euc is NOT set)
        -1: Distance exceeds the maximum distance (-max)
        -2: First atom is solvent-inaccessible
        -3: Second atom is solvent-inaccessible
        -4: Both atoms are solvent-inaccessible
        -5: First atom is in a cavity which prohibited proper shortest path calculations


Virtual cross-links are sorted first by decreasing probability, then by increasing SAS distance and finally by increasing Euclidean distance.

Commandline PARAMETER:
INPUT/OUTPUT:
        -infile <path>  Any PDB file; .tar, .gz and .tar.gz files with PDB file content are also accepted [required].
        -xSC    [switch]        Removes only side chain atoms of cross-linked amino acids except for CB atoms and keeps -radius at 1.4 prior to calculating SAS distances. This might be of value when side chain conformations of cross-linked residues are unknown [optional].
        -bb     [switch]        Reads in only backbone and beta carbon atom coordinates from the input file and increases -radius to 2.0. Be cautious using the option, as it might cause shortest path calculalations through "molecular tunnels" in your protein [optional][see also -xSC].
        -dist   <path>  Distance file holding at least the first 4 columns of the Xwalk output format. The file will be used to extract the indices and the residue pairs for the distance calculation [optional].
        -keepName       [switch]        Uses the same name (2nd column) in the output as in the distance file. [optional].
        -out    <path>  Writes output to this file, otherwise output is directed to the STDOUT channel. If -pymol is set than filename must have .pml filename ending [optional].
        -f      [switch]        Forces output to be written into a file even if file already exists [optional].
        -pymol  [switch]        Outputs a PyMOL (http://www.pymol.org/) script highlighting the calculated distances of the potential cross-links [optional].
        -v      [switch]        Outputs various information other than distances [optional].
        -grid   [switch]        Outputs on STDOUT channel the grid, which is used to calculate the Solvent Accessible Surface Distance. The grid is in PDB format with distances in the B-factor column [optional].

RESIDUE/ATOM SELECTION:
        -aa1    [String]        Three letter code of 1st amino acid. To specify more than one amino acid use '#' as a delimeter [required, if -r1 is not set].
        -aa2    [String]        Three letter code of 2nd amino acid. To specify more than one amino acid use '#' as a delimeter [required, if -r2 is not set].
        -r1     [String]        Amino acid residue number. To specify more than one residue number use '#' as a delimeter. [required, if -aa1 is not set].
        -r2     [String]        Amino acid residue number. To specify more than one residue number use '#' as a delimeter. [required, if -aa2 is not set].
        -c1     [String]        Chain ids for -aa1 or -r1. For blank chain Id use '_'. To specify more than one chain Id, append chain ids to a single string, e.g. ABC [optional](default: all chain Ids).
        -c2     [String]        Chain ids for -aa2 or -r2. For blank chain Id use '_'. To specify more than one chain Id, append chain ids to a single string, e.g. ABC [optional](default: all chain Ids).
        -a1     [String]        Atom type for -aa1 or -r1. To specify more than one atom type use '#' as a delimeter. [optional].
        -a2     [String]        Atom type for -aa2 or -r2. To specify more than one atom type use '#' as a delimeter. [optional].
        -l1     [String]        Alternative location id for -aa1 or -r1. To specify more than one alternative location, append alternative location ids to a single string, e.g. AB [optional].
        -l2     [String]        Alternative location id for -aa2 or -r1. To specify more than one alternative location, append alternative location ids to a single string, e.g. AB [optional].
        -intra  [switch]        Outputs only "intra-molecular" distances [optional].
        -inter  [switch]        Outputs only "inter-molecular" distances [optional].
        -homo   [double]        Outputs only shortest distance of potential cross-links between equally numbered residues. Reduces redundancy if PDB file is a homomeric protein complex. [optional].

DIGESTION RELATED:
        -trypsin        [switch]        Digests in silico the protein with trypsin and excludes peptides that are shorter than 5 AA or larger than 40 AA [optional].

DISTANCE RELATED:
        -max    [double]        Calculates distances in Angstroem only up-to this value, where the value must be smaller than 100.0 for SAS distance calculations. (default: 34.0).
        -euc    [switch]        Skips Solvent-Path-Distance calculation and outputs only Euclidean distances [optional].
        -prob   [switch]        Outputs probability information for each vXL as determined by experimental data on DSS and BS3 cross-linking experiments [optional].
        -bfactor        [switch]        Adds the uncertainty of the atom coordinates as expressed by their B-factor/temperature factor to the maximum distance threshold [optional].

SOLVENT-PATH-DISTANCE GRID RELATED:
        -radius [double]        Solvent radius for calculating the solvent accessible surface area [optional](default 1.4).
        -space  [double]        Spacing in Angstroem between grid cells. [optional](default 1.0).

"""

import os
import re

import numpy as np
import pandas as pd


def extract_xwalk_cmd(xwalk_cmd):
    aa1 = re.findall('-aa1 (.+?) ', xwalk_cmd, re.I)[0]
    aa2 = re.findall('-aa2 (.+?) ', xwalk_cmd, re.I)[0]
    a1 = re.findall('-a1 (.+?) ', xwalk_cmd, re.I)[0]
    a2 = re.findall('-a2 (.+?) ', xwalk_cmd, re.I)[0]
    inter_intra = re.findall('-(inter|intra)', xwalk_cmd, re.I)
    inter_intra = inter_intra[0] if inter_intra else None
    max_length = re.findall('-max (\d+?)[> ]', xwalk_cmd, re.I)[0]
    return aa1, aa2, a1, a2, inter_intra, max_length


def read_command_file(cmd_file, pdb_name):
    cmd_list = []
    with open(os.path.abspath(cmd_file), 'r') as cmd_handle:
        for each_line in cmd_handle:
            each_cmd = each_line.strip('\n')
            if not each_cmd:
                continue
            aa1, aa2, a1, a2, inter_intra, maxlength = extract_xwalk_cmd(each_cmd)

            rearranged_cmd = 'java -Xmx1024m Xwalk -infile {}.pdb -aa1 {} -aa2 {} -a1 {} -a2 {}{} -max {} -bb >'.format(
                pdb_name, aa1, aa2, a1, a2, ' -{}'.format(inter_intra) if inter_intra else '', maxlength)
            out_filename = '{pdb_filename}-{aa1}_{aa2}_{a1}_{a2}{inter_intra}-{maxlength}.txt'.format(
                pdb_filename=pdb_name, aa1=aa1, aa2=aa2, a1=a1, a2=a2,
                inter_intra='-{}'.format(inter_intra) if inter_intra else '', maxlength=maxlength)

            cmd_list.append((rearranged_cmd, out_filename))
    return cmd_list


def xwalk_run(cmd_list, result_add):
    cmd_num = len(cmd_list)
    for cmd_series, _ in enumerate(cmd_list):
        rearranged_cmd, out_filename = _
        each_cmd = '{}"{}"'.format(rearranged_cmd, os.path.join(result_add, out_filename))
        print('{}/{} Now running {}'.format(cmd_series + 1, cmd_num, each_cmd))
        os.system(each_cmd)


def merge_result(result_add):
    result_file_list = os.listdir(result_add)
    file_num = len(result_file_list)
    with open(os.path.join(result_add, 'MergedIntra.txt'), 'w') as intra_handle, open(os.path.join(result_add, 'MergedInter.txt'), 'w') as inter_handle:
        for file_series, each_result in enumerate(result_file_list):
            print('{}/{} Merging {}'.format(file_series + 1, file_num, each_result))
            intra_handle.write(each_result + '\n')
            inter_handle.write(each_result + '\n')

            result_path = os.path.join(result_add, each_result)
            with open(result_path, 'r') as result_handle:
                for each_line in result_handle:
                    if not each_line.strip('\n'):
                        continue
                    split_line = each_line.split('\t')
                    first_site = split_line[2]
                    second_site = split_line[3]
                    if first_site.split('-')[2] == second_site.split('-')[2]:
                        intra_handle.write(each_line)
                    else:
                        inter_handle.write(each_line)
            intra_handle.write('\n')
            inter_handle.write('\n')


class SiteConverter(object):

    def __init__(
            self,
            conversion_file,
            sheet_name=None,
            pdb_sites_col_name='PDBPos',
            fasta_sites_col_name='FastaPos',
            aa3_col_name='',
    ):

        self.conversion_info_df = self._read_conversion_file(conversion_file, sheet_name)

        self.pdb_sites = self.conversion_info_df[pdb_sites_col_name].tolist()
        self.fasta_sites = self.conversion_info_df[fasta_sites_col_name].tolist()
        self.correspond_aa3 = self.conversion_info_df[aa3_col_name].tolist()

    @staticmethod
    def _read_conversion_file(conversion_file, sheet_name=None) -> pd.DataFrame:
        if conversion_file.endswith('xlsx') or conversion_file.endswith('xls'):
            df = pd.read_excel(conversion_file, dtype=str, sheet_name=sheet_name)
        elif conversion_file.endswith('txt') or conversion_file.endswith('tsv'):
            df = pd.read_csv(conversion_file, sep='\t', dtype=str)
        elif conversion_file.endswith('csv'):
            df = pd.read_csv(conversion_file, sep=',', dtype=str)
        else:
            raise
        return df

    def pdb_to_fasta(self, pdb_pos):
        try:
            idx = self.pdb_sites.index(pdb_pos)
            return self.correspond_aa3[idx], self.fasta_sites[idx]
        except ValueError:
            print('not find pdb position: {}'.format(pdb_pos))
            return None

    def fasta_to_pdb(self, fasta_pos):
        """
        :param fasta_pos: type=str
        :return: similar to previous function
        """
        try:
            idx = self.fasta_sites.index(fasta_pos)
            return self.correspond_aa3[idx], self.pdb_sites[idx]
        except ValueError:
            print('not find fasta position: {}'.format(fasta_pos))
            return None

    def pdb_to_aa(self, pdb_pos):
        try:
            idx = self.pdb_sites.index(pdb_pos)
            return self.correspond_aa3[idx]
        except ValueError:
            print('not find pdb position {}'.format(pdb_pos))
            return None

    def fasta_to_aa(self, fasta_pos):
        try:
            idx = self.fasta_sites.index(fasta_pos)
            return self.correspond_aa3[idx]
        except ValueError:
            print('not find fasta position {}'.format(fasta_pos))
            return None


class XwalkRunner(object):
    """

    """
    xwalk_command_template = ('java -Xmx1024m -cp "{xwalk_bin}" Xwalk'
                              ' -infile "{pdb_path}"'
                              ' -pymol -out "{pymol_output}"'
                              ' -dist "{xwalk_input_file}"'
                              ' -max {max_distance} -bb -homo >"{output_path}"')

    def __init__(
            self,
            result_df: pd.DataFrame,
            link_info_colname,
            xwalk_dist_folder,
            pymol_folder,
            xwalk_bin_folder,
            pdb_path,
            max_distance: int = 100,
            equal_struct_fasta_aa_pos: bool = False,
            site_converter: dict = None,
            prot_mapper=None,
            linker_aa_file: str = None,
            task_identifier: str = None,
            logger=None,
    ):

        self.result_df = result_df
        self.link_info_colname = link_info_colname
        self.all_link_info = self.result_df[self.link_info_colname].drop_duplicates().tolist()

        self.xwalk_dist_folder = xwalk_dist_folder
        self.pymol_folder = pymol_folder
        self.xwalk_bin_folder = xwalk_bin_folder
        self.pdb_path = pdb_path
        for folder in [xwalk_dist_folder, pymol_folder, xwalk_bin_folder]:
            os.makedirs(folder, exist_ok=True)

        self.max_distance = max_distance
        if equal_struct_fasta_aa_pos:
            self.use_site_converter = False
        else:
            self.use_site_converter = True
            if site_converter is None:
                raise ValueError
            self.site_converter = site_converter

        if prot_mapper is None:
            self.prot_mapper_to_pdb = {
                'DNGAS': 'A',
                'DNG': 'A',
                'A': 'A',
                'GNB1': 'B',
                'B': 'B',
                'GNG2': 'G',
                'G': 'G',
                'Nb35': 'N',
                'NB35': 'N',
                'N': 'N',
                'GLP1R': 'R',
                'R': 'R',
                'GLP1': 'P',
                'GL': 'P',
                'P': 'P',
            }
        else:
            self.prot_mapper_to_pdb = prot_mapper

        if linker_aa_file is None:
            self.linker_linked_aa = {
                'ArGO': (('Arg', 'Arg'),),
                'KArGO': (('Lys', 'Arg'),),
                'EGS': (('Lys', 'Lys'),),
                'DSS': (('Lys', 'Lys'),),
                'BS3': (('Lys', 'Lys'),),
                'PDH': (('Glu', 'Asp'), ('Glu', 'Glu'), ('Asp', 'Asp')),
                'EDC': (('Lys', 'Glu'), ('Lys', 'Asp')),
            }
        else:
            self.linker_linked_aa = dict()
            self._read_linker_linking_info(linker_aa_file)

        self.task_identifier = task_identifier
        self.logger = logger
        self.xwalk_result = dict()

        self.output_file_path = os.path.join(self.xwalk_dist_folder, self.task_identifier + '-xwalk_output.tsv')

    def _read_linker_linking_info(self, linker_aa_file):
        with open(linker_aa_file, 'r') as f:
            linker_info = f.read().split('\n')
        for row in linker_info:
            if 'linker' in row:
                continue
            elif not row:
                continue
            else:
                split_row = row.split(',')
                if len(split_row) <= 1:
                    continue
                self.linker_linked_aa[split_row[0]] = (
                    (split_row[aa_num * 2], split_row[aa_num * 2 + 1])
                    for aa_num in range(len(split_row[1:]) // 2)
                )

    def _convert_sites(self, site, consensus_prot):
        if self.use_site_converter:
            convert_result = self.site_converter[consensus_prot].fasta_to_pdb(site)
            if convert_result is None:
                return None
            aa3, pdb_pos = convert_result
            return '{}-{}-{}-CA'.format(aa3, pdb_pos, consensus_prot)
        else:
            convert_result = self.site_converter[consensus_prot].pdb_to_aa(site)
            if convert_result is None:
                return None
            aa3 = convert_result
            return '{}-{}-{}-CA'.format(aa3, site, consensus_prot)

    def generate_xwalk_input(self):
        if self.logger is not None:
            self.logger.info('Generate xwalk input for {}'.format(self.task_identifier))

        xwalk_input = ''
        for idx, link_info in enumerate(self.all_link_info, 1):
            re_match = re.match('(.+?)\((\d+)\)-(.+?)\((\d+)\).*', link_info)

            input_aa1 = self._convert_sites(re_match.group(2), self.prot_mapper_to_pdb[re_match.group(1)])
            input_aa2 = self._convert_sites(re_match.group(4), self.prot_mapper_to_pdb[re_match.group(3)])

            if input_aa1 is not None and input_aa2 is not None:
                xwalk_input += '{}\t{}\t{}\t{}\n'.format(
                    idx, os.path.basename(self.pdb_path),
                    input_aa1,
                    input_aa2
                )
        with open(os.path.join(self.xwalk_dist_folder, self.task_identifier + '.tsv'), 'w') as f:
            f.write(xwalk_input)

    def fillin_command(self, start_time):
        str_pymol_filename = self.task_identifier + start_time + '.pml'
        command = self.xwalk_command_template.format(
            xwalk_bin=self.xwalk_bin_folder,
            pdb_path=self.pdb_path,
            pymol_output=os.path.join(self.pymol_folder, str_pymol_filename),
            xwalk_input_file=os.path.join(self.xwalk_dist_folder, self.task_identifier + '.tsv'),
            max_distance=self.max_distance,
            output_path=self.output_file_path,
        )
        return command

    def run_xwalk(self, start_time=None):
        if start_time is None:
            start_time = ''
        command = self.fillin_command(start_time)
        print('start cmd {}'.format(command))
        os.system(command)
        print('{} complete'.format(self.task_identifier))

    def collect_xwalk_result(self, output_file_path=None):
        if output_file_path is not None:
            self.output_file_path = output_file_path

        with open(self.output_file_path, 'r') as f:
            xwalk_output = f.read().strip('\n').split('\n')

        for row in xwalk_output:
            split_row = row.split('\t')
            idx = int(split_row[0])
            surface_distance = float(split_row[6])
            link_info = self.all_link_info[idx - 1]
            self.xwalk_result[link_info] = surface_distance

    def add_distance_to_plink_result(self):
        self.result_df['SASDs'] = self.result_df[self.link_info_colname].apply(
            lambda x: self.xwalk_result[x] if x in self.xwalk_result else np.nan)

    def get_result(self):
        return self.result_df.copy()

    def run_xwalk_pipeline(self):
        self.generate_xwalk_input()
        self.run_xwalk()
        self.collect_xwalk_result()
        self.add_distance_to_plink_result()
        return self.get_result()



"""

"""


class ModComp:
    Mods = {
        'Carbamidomethyl (C)': {'C': 2, 'N': 1, 'O': 1, 'H': 3},
        'Oxidation (M)': {'O': 1},
        'Phospho (STY)': {'H': 1, 'P': 1, 'O': 3},
        'Acetyl (Protein N-term)': {'C': 2, 'O': 1, 'H': 2},
        'Deamidation (NQ)': {'O': 1, 'N': -1, 'H': -1},
        'Methyl (KR)': {'C': 1, 'H': 2},
        'Dimethyl (KR)': {'C': 2, 'H': 4},

    }

    NeutralLoss = {
        'H2O': {'H': 2, 'O': 1}
    }


class BasicModification:
    # Pre-defined mods, where key means query mod and value is used for assembling queried mod
    StandardMods = {
        'Carbamidomethyl': 'Carbamidomethyl',
        'Oxidation': 'Oxidation',
        'Phospho': 'Phospho',
        'Acetyl': 'Acetyl'
    }

    # To extend the query space. Each mod has its alias and itself for quering
    __ModAliasList = {
        'Carbamidomethyl': ['Carbamidomethyl', 'Carbamid', 'Carb', 'Carbamidomethyl[C]'],
        'Oxidation': ['Oxidation', 'Oxi', 'Ox', 'Oxidation[M]'],
        'Phospho': ['Phospholation', 'Phospho', 'Phos', ],
        'Acetyl': ['_[Acetyl (Protein N-term)]'],
    }
    ModAliasDict = {}
    for standard, aliases in __ModAliasList.items():
        for alias in aliases:
            ModAliasDict[alias] = standard
    for alias in list(ModAliasDict.keys()):
        ModAliasDict[alias.upper()] = ModAliasDict[alias]
        ModAliasDict[alias.lower()] = ModAliasDict[alias]

    # Mod rule. This defines the method for mod assembly
    StandardModRule = r'[{mod} ({aa})]'
    ModRuleDict = {'standard': StandardModRule,
                   }

    ModAA = {
        'Carbamidomethyl': ['C'],
        'Oxidation': ['M'],
        'Phospho': ['S', 'T', 'Y'],
    }

    ExtendModAA = ModAA.copy()
    ExtendModAA['Phospho'].extend(['H', 'R', 'K', 'D', 'G', 'M', 'V', 'P', 'N', 'A'])
